#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
AutoTrader Help
"""

import pyfiglet

def usage():
    """ Print usage options. """
    banner = pyfiglet.figlet_format("AUTOTRADER")
    print(banner)
    
    print("AutoTrader is an algorithmic trading development platform.")
    print("\nIt has three run modes:")
    print("  1. Backtest mode")
    print("  2. Livetrade mode")
    print("  3. Scan mode")
    print("By default, AutoTrader will run in livetrade mode.\n")
    
    print("The user options are shown below.")

    print("--------------------------------------------------------------" \
          + "---------------")
    print("Attribute                           Comment")
    print("--------------------------------------------------------------" \
          + "---------------")
    print("Required:")
    print("  config (str)                       filename of config file")
    print("\nOptional:")
    print("  help                               show help for usage")
    print("  verbosity (int)                    set verbosity (0,1,2)")
    print("  backtest                           run in backtesting mode")
    print("  plot                               plot results of backtest")
    print("  notify (int)                       notify by email when ordering")
    print("  log                                log backtest results to file")
    print("  analyse                            run correlation study of indicators")
    print("  scan                               run in scan mode only")
    print("  optimise                           optimise strategy parameters")
    print("  instruments                        specify specific instruments")
    print("  data                               load custom price data file")
    print("")


def option_help(option):
    ''' Print usage instructions. '''
    
    banner = pyfiglet.figlet_format("AUTOTRADER")
    print(banner)
    if option == 'config' or option == 'c':
        print("Help for 'config' option:")
        print("-----------------------------------")
        print("A configuration file must be specified to run AutoTrader. The")
        print("file must be written as a .yaml file according to the template")
        print("provided in the config/ directory.")
        
        print("Note that the file extension should not be included. The full")
        print("file path does not need to be specified, AutoTrader will search")
        print("for the file in the config/ directory.")

        
    elif option == 'verbosity' or option == 'v':
        print("Help for 'verbosity' option:")
        print("-----------------------------------")
        print("The verbosity flag is used to set the level of output")
        print("displayed by the code. A verbosity of zero supresses all")
        print("output, while a value greater than zero will show more details")
        print("of what the code is doing.")
        print("Verbosity settings will not affect error output.")
        
        print("\nDefault value: 0")

        
    elif option == 'backtest' or option == 'b':
        print("Help for 'backtest' option:")
        print("-----------------------------------")
        print("The backtest flag is used to run the strategy in backtest")
        print("mode.")
        
        print("\nDefault value: False")
 
        
    elif option == 'plot' or option == 'p':
        print("Help for 'plot' option:")
        print("-----------------------------------")
        print("The plot option is used to create a plot of the price chart")
        print("and strategy-specific indicators. It may be used for both")
        print("livetrading and backtesting.")
        
        print("\nDefault value: False")

        
    elif option == 'notify' or option == 'n':
        print("Help for 'notify' option:")
        print("-----------------------------------")
        print("The notify option may be used to enable email notifications")
        print("of livetrade activity and AutoScan results.")
        
        print("Options:")
        print("  -n 0: No emails will be sent.")
        print("  -n 1: Minimal emails will be sent (summaries only).")
        print("  -n 2: All emails will be sent (every order and summary).")
        
        print("Note: if daily email summaries are desired, email_manager must")
        print("be employed in another scheduled job to send the summary.")
        
        print("\nDefault value: 0")

        
    elif option == 'log' or option == 'l':
        print("Help for 'log' option:")
        print("-----------------------------------")
        print("The log option allows logging of backtest results to a")
        print("logfile. The log file will be written to logfiles/ and")
        print("includes key statistics of the backtest, such as win rate,")
        print("number of trades and profit. The configuration file is also")
        print("embeded in the log file for future reference.")
        
        print("\nDefault value: False")

        
    elif option == 'analyse' or option == 'a':
        print("Help for 'analyse' option:")
        print("-----------------------------------")
        print("Analyser. More information coming soon.")
        
        print("\nDefault value: False")

        
    elif option == 'scan' or option == 's':
        print("Help for 'scan' option:")
        print("-----------------------------------")
        print("Automated market scanner. When running AutoTrader in this mode,")
        print("the market will be scanned for entry conditions based on the")
        print("strategy in the configuration file.")
        print("When the notify flag is included, an email will be sent to")
        print("notify the recipients in the email list of the signal.")
        print("This option requires an index or instrument to scan as an")
        print("input.")
        
        print("Note: if email notifications are enabled and there are no scan")
        print("hits, no email will be sent. However, if you still wish to receive")
        print("emails regardless, set the verbosity of the code to 2. In this")
        print("case, an email will be sent on the completion of each scan,")
        print("regardless of the results.")
        
        print("\nDefault value: False")

        
    elif option == 'optimise' or option == 'o':
        print("Help for 'optimise' option:")
        print("-----------------------------------")
        print("When this flag is included, AutoTrader will return a dictionary")
        print("containing backtest results, to be used by the optimiser.")
        print("This option is to be used internally by AutoOptimise.")
        
        print("\nDefault value: False")

    
    elif option == 'data' or option == 'd':
        print("Help for 'data' option:")
        print("-----------------------------------")
        print("This flag may be used to specify the filename for custom")
        print("historical price data. Note that if this flag is included,")
        print("the backtesting times specified in the config file will no")
        print("longer be used.")
        
        print("Currently, data must be located in the price data directory")
        print("to be used.")
        
        print("Important: if a data file is provided, this will also be used")
        print("be used for the quote data. That is, currency conversions will")
        print("not be accounted for.")
        
        print("\nDefault value: None")
        
    
    elif option == 'instruments' or option == 'i':
        print("Help for 'instruments' option:")
        print("-----------------------------------")
        print("This flag may be used to specify instruments to run AutoTrader")
        print("on, overwriting the watchlist in the strategy config file.")
        
        print("\nDefault value: None")

        
    elif option == "general":
        print("General help.")
        print("")
        
    else:
        print("Unrecognised flag ({}).".format(option))
    
    if option != "general":
        print("\n\nFor general help, pass 'general' to help.\n")
    

def stream_help(option=None):
    """ 
    Print help for autostream.
    """
    
    if option == 'usage' or option is None:
        banner = pyfiglet.figlet_format("AUTOSTREAM")
        print(banner)
        print("Utility to stream price data and write to text file.")
        print("")
        print("--------------------------------------------------------------" \
              + "---------------")
        print("Flag                                 Comment [short flag]")
        print("--------------------------------------------------------------" \
              + "---------------")
        print("Required:") 
        print("  --instrument 'XXX_YYY'             instrument to stream [-i]")
        print("  --granularity 'M15'                candlestick granularity [-g]")
        print("\nOptional:")
        print("  --help                             show help for usage [-h]")
        print("  --verbosity <int>                  set verbosity (0,1,2) [-v]")
        print("  --max-candles <10>                 max number of candles to store [-N]")
        print("  --index ''                         specify index to stream [-I]")
        print("")
        print("Note: if multiple instruments are requested, they must be entered")
        print("as comma separated text with no spaces. Example:")
        print("-i EUR_USD,USD_JPY,AUD_CAD")

    elif option == 'instrument' or option == 'i':
        print("Help for '--instrument' (-c) option:")
        
        print("\nExample usage:")
        print("./AutoStream.py -c my_config_file")
        
    elif option == 'verbosity' or option == 'v':
        print("Help for '--verbosity' (-v) option:")
        print("-----------------------------------")
        print("The verbosity flag is used to set the level of output.")
    
    elif option == 'index' or option == 'I':
        print("Help for '--verbosity' (-v) option:")
        print("-----------------------------------")
        print("Specify an index to stream. ")
        print("This flag takes precedence over -i if both are specified.")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import importlib
import time
import pytz
import pandas as pd
import numpy as np
import threading
from shutil import copy2
from datetime import datetime
from autotrader.emailing import emailing
from autotrader.lib import autodata, environment_manager
from autotrader.lib.read_yaml import read_yaml
from autotrader.autostream import AutoStream


class AutoTraderBot():
    '''
    AutoTrader Bot.
    ---------------
    
    Attributes
    ----------
    broker : class
        The broker class instance.
        
    instrument : str
        The instrument being traded by the bot.
    
    strategy : class
         The strategy being traded by the bot.
    

    Methods
    -------
    update(i):
        Update strategy with latest data and generate latest signal.
    
    '''
    
    def __init__(self, instrument, strategy_config, broker, autotrader_attributes):
        '''
        AutoTrader Bot initialisation. 
        '''

        # Inherit user options from autotrader
        self.home_dir           = autotrader_attributes.home_dir
        self.scan_mode          = autotrader_attributes.scan_mode
        self.scan_index         = autotrader_attributes.scan_index
        self.scan_results       = {}
        self.broker_utils       = autotrader_attributes.broker_utils
        self.email_params       = autotrader_attributes.email_params
        self.notify             = autotrader_attributes.notify
        self.verbosity          = autotrader_attributes.verbosity
        self.order_summary_fp   = autotrader_attributes.order_summary_fp
        self.backtest_mode      = autotrader_attributes.backtest_mode
        self.data_start         = autotrader_attributes.data_start
        self.data_end           = autotrader_attributes.data_end
        self.base_currency      = autotrader_attributes.backtest_base_currency
        self.environment        = autotrader_attributes.environment
        self.feed               = autotrader_attributes.feed
        self.data_file          = autotrader_attributes.data_file
        self.optimise_mode      = autotrader_attributes.optimise_mode
        self.check_data_alignment = autotrader_attributes.check_data_alignment
        self.allow_dancing_bears = autotrader_attributes.allow_dancing_bears
        self.use_stream         = autotrader_attributes.use_stream
        self.stream_config      = autotrader_attributes.stream_config
        self.MTF_initialisation = autotrader_attributes.MTF_initialisation
        
        # Assign local attributes
        self.instrument         = instrument
        self.broker             = broker
        
        # Unpack strategy parameters and assign to strategy_params
        interval                = strategy_config["INTERVAL"]
        period                  = strategy_config["PERIOD"]
        risk_pc                 = strategy_config["RISK_PC"] if 'RISK_PC' in strategy_config else 0
        sizing                  = strategy_config["SIZING"] if 'SIZING' in strategy_config else 0
        params                  = strategy_config["PARAMETERS"]
        strategy_params                 = params
        strategy_params['granularity']  = interval
        strategy_params['risk_pc']      = risk_pc
        strategy_params['sizing']       = sizing
        strategy_params['period']       = period
        self.strategy_params            = strategy_params
        
        # Import Strategy
        strat_module            = strategy_config["MODULE"]
        strat_name              = strategy_config["CLASS"]
        strat_package_path      = os.path.join(self.home_dir, "strategies") 
        strat_module_path       = os.path.join(strat_package_path, strat_module) + '.py'
        strat_spec              = importlib.util.spec_from_file_location(strat_module, strat_module_path)
        strategy_module         = importlib.util.module_from_spec(strat_spec)
        strat_spec.loader.exec_module(strategy_module)
        strategy                = getattr(strategy_module, strat_name)
        
        # Get broker configuration 
        global_config_fp = os.path.join(self.home_dir, 'config', 'GLOBAL.yaml')
        if os.path.isfile(global_config_fp):
            global_config = read_yaml(global_config_fp)
        else:
            global_config = None
        broker_config  = environment_manager.get_config(self.environment,
                                                        global_config,
                                                        self.feed)
   
        # Start price streaming
        if self.use_stream and self.backtest_mode is False:
            
            # Check how many granularities were requested
            if len(interval.split(',')) > 1:
                # MTF strategy
                self.base_interval = interval.split(',')[0]
                self.MTF_intervals = interval.split(',')[1:]
                
                # Initiate time_to_download dict
                self.time_to_download = {}
                for granularity in self.MTF_intervals:
                    self.time_to_download[granularity] = datetime.now(tz=pytz.utc)
                
            else:
                # Single timeframe strategy
                self.base_interval = interval
                self.MTF_intervals = []
            
            # Start stream
            self._initiate_stream()

        # Multiple time-frame initialisation option
        if self.MTF_initialisation:
            # Only retrieve MTF_data once upon initialisation and store
            # instantiation MTF_data
            self.MTF_data = None
        
        # Data retrieval
        self.get_data = autodata.GetData(broker_config, self.allow_dancing_bears)
        data, quote_data, MTF_data = self._retrieve_data(instrument, self.feed)
        
        # Data assignment
        if MTF_data is None:
            strat_data = data
        else:
            strat_data = MTF_data
        
        # Instantiate Strategy
        include_broker = strategy_config['INCLUDE_BROKER'] if 'INCLUDE_BROKER' in strategy_config else False
        if include_broker:
            my_strat = strategy(params, strat_data, instrument, self.broker, self.broker_utils)
        else:
            my_strat = strategy(params, strat_data, instrument)
            
        # Assign strategy to local attributes
        self.strategy           = my_strat
        self.data               = data
        self.quote_data         = quote_data
        self.latest_orders      = []
        
        # Assign strategy attributes for tick-based strategy development
        if self.backtest_mode:
            self.strategy._backtesting = True
        if interval.split(',')[0] == 'tick':
            self.strategy._tick_data = True
        
        if int(self.verbosity) > 0:
                print("AutoTraderBot assigned to analyse {}".format(instrument),
                      "on {} timeframe using {}.\n".format(self.strategy_params['granularity'],
                                                         strategy_config['NAME']))
    
    
    def _initiate_stream(self):
        '''
        Spawns AutoStream into a new thread.
        '''
        
        record_ticks = False
        record_candles = False
        if self.base_interval == 'tick' or self.base_interval == 'ticks':
            record_ticks = True
        else:
            record_candles = True
        
        
        stream_granularity = self.base_interval
        self.no_candles = self.strategy_params['period']
        
        self.AS = AutoStream(self.home_dir, 
                             self.stream_config, 
                             self.instrument, 
                             granularity = stream_granularity,
                             record_ticks = record_ticks, 
                             record_candles = record_candles,
                             no_candles = self.no_candles,
                             bot = self)
        
        stream_thread = threading.Thread(target = self.AS.start, 
                                         args=(), daemon=False)
        print('Spawning new thread to stream data.')
        stream_thread.start()
        
    
    def _recieve_stream_data(self):
        '''
        Method to tell AutoStream to send data to bot. Called from bot manager.
        '''
        
        self.AS.update_bot = True
    
    def _update_strategy_data(self, data=None):
        '''
        Method to update strategy with latest data. Called by the bot manager
        and autostream.
        '''
        
        if data is not None:
            # Update data attribute (for livetrade compatibility)
            self.data = data
        
        # Retrieve new data
        new_data, _, MTF_data = self._retrieve_data(self.instrument, 
                                                    self.feed,
                                                    base_data = data)
        
        # Check for MTF_data
        if MTF_data is None:
            strat_data = new_data
        else:
            strat_data = MTF_data
        
        # Update strategy with new data
        self.strategy.initialise_strategy(strat_data)
        
    
    def _retrieve_data(self, instrument, feed, base_data = None):
        '''
        Retrieves price data from AutoData.
        '''
        
        interval    = self.strategy_params['granularity']
        period      = self.strategy_params['period']
        price_data_path = os.path.join(self.home_dir, 'price_data')
        
        if self.backtest_mode is True:
            ' ~~~~~~~~~~~~~~~~~~~ Running in backtest mode ~~~~~~~~~~~~~~~~~~ '
            self.get_data.base_currency = self.base_currency
            
            from_date       = self.data_start
            to_date         = self.data_end
            
            if self.data_file is not None:
                custom_data_file        = self.data_file
                custom_data_filepath    = os.path.join(price_data_path,
                                                       custom_data_file)
                if int(self.verbosity) > 1:
                    print("Using data file specified ({}).".format(custom_data_file))
                data            = pd.read_csv(custom_data_filepath, 
                                              index_col = 0)
                data.index = pd.to_datetime(data.index)
                quote_data = data
                
            else:
                if int(self.verbosity) > 1:
                    print("\nDownloading OHLC price data for {}.".format(instrument))
                
                if self.optimise_mode is True:
                    # Check if historical data already exists
                    historical_data_name = 'hist_{0}{1}.csv'.format(interval, instrument)
                    historical_quote_data_name = 'hist_{0}{1}_quote.csv'.format(interval, instrument)
                    data_dir_path = os.path.join(self.home_dir, 'price_data')
                    historical_data_file_path = os.path.join(self.home_dir, 
                                                             'price_data',
                                                             historical_data_name)
                    historical_quote_data_file_path = os.path.join(self.home_dir, 
                                                             'price_data',
                                                             historical_quote_data_name)
                    
                    if not os.path.exists(historical_data_file_path):
                        # Data file does not yet exist
                        data        = getattr(self.get_data, feed.lower())(instrument,
                                                         granularity = interval,
                                                         start_time = from_date,
                                                         end_time = to_date)
                        quote_data  = getattr(self.get_data, feed.lower() + '_quote_data')(data,
                                                                                      instrument,
                                                                                      interval,
                                                                                      from_date,
                                                                                      to_date)
                        data, quote_data    = self.broker_utils.check_dataframes(data, quote_data)
                        
                        # Check if price_data folder exists
                        if not os.path.exists(data_dir_path):
                            # If price data directory doesn't exist, make it
                            os.makedirs(data_dir_path)
                            
                        # Save data in file/s
                        data.to_csv(historical_data_file_path)
                        quote_data.to_csv(historical_quote_data_file_path)
                        
                    else:
                        # Data file does exist, import it as dataframe
                        data = pd.read_csv(historical_data_file_path, 
                                           index_col = 0)
                        quote_data = pd.read_csv(historical_quote_data_file_path, 
                                                 index_col = 0)
                    
                    # TODO - add support of MTF for optimisation
                    MTF_data = None
                        
                else:
                    # Running in single backtest mode
                    MTF_data = {}
                    for granularity in interval.split(','):
                        data        = getattr(self.get_data, feed.lower())(instrument,
                                                             granularity = granularity,
                                                             start_time = from_date,
                                                             end_time = to_date)
                        
                        # Only get quote data for first granularity
                        if granularity == interval.split(',')[0]:
                            quote_data  = getattr(self.get_data, feed.lower() + '_quote_data')(data,
                                                                            instrument,
                                                                            granularity,
                                                                            from_date,
                                                                            to_date)
                        
                            data, quote_data    = self.broker_utils.check_dataframes(data, quote_data)
                        
                        MTF_data[granularity] = data
                    
                    # Extract first dataset to use as base
                    first_granularity = interval.split(',')[0]
                    data = MTF_data[first_granularity]
                    quote_data = quote_data
                
                if len(MTF_data) == 1:
                    MTF_data = None
                
                if int(self.verbosity) > 1:
                    print("  Done.\n")
            
            return data, quote_data, MTF_data
        
        else:
            ' ~~~~~~~~~~ Running in livetrade mode or scan mode ~~~~~~~~~~~~~ '
            
            if self.use_stream:
                # Streaming data
                
                # First assign data
                if base_data is not None:
                    data = base_data
                else:
                    print("Stream data has not been recieved yet.")
                    print("Passing NoneType as data.")
                    data = None
                
                # Now retrieve MTF data
                if len(interval.split(',')) > 1:
                    # Fetch MTF data
                    MTF_data = {self.base_interval: data}
                    
                    if self.MTF_initialisation:
                        # Download MTF data for strategy initialisation only
                        if self.MTF_data is None:
                            # MTF_data has not been retrieved yet
                            for granularity in self.MTF_intervals:
                                data = getattr(self.get_data, feed.lower())(instrument,
                                                                            granularity = granularity,
                                                                            count=period)
                                
                                if self.check_data_alignment:
                                    data = self._verify_data_alignment(data, instrument, feed, period, 
                                                                       price_data_path)
                                
                                MTF_data[granularity] = data
                            
                            self.MTF_data = MTF_data
                            
                        else:
                            # MTF_data already exists, reuse
                            MTF_data = self.MTF_data
                            
                            # Replace base_interval data with latest data
                            MTF_data[self.base_interval] = data
                            
                    else:
                        # Download MTF data each update
                        for granularity in self.MTF_intervals:
                            if datetime.now(tz=pytz.utc) > self.time_to_download[granularity]:
                                # Update MTF data
                                data = getattr(self.get_data, feed.lower())(instrument,
                                                                            granularity = granularity,
                                                                            count = period)
                            
                                if self.check_data_alignment:
                                    data = self._verify_data_alignment(data, instrument, feed, period, 
                                                                       price_data_path)
                                
                                # Append to MTF_data
                                MTF_data[granularity] = data
                                
                                # Update next time_to_download
                                self.time_to_download[granularity] = self._next_candle_open(granularity)
                            
                            else:
                                # Use previously downloaded MTF data
                                MTF_data[granularity] = self.MTF_data[granularity]
                        
                        # Update self.MTF_data with latest MTF data
                        self.MTF_data = MTF_data
                                
                            
                else:
                    # There is only one timeframe of data
                    MTF_data = None
                
            
            elif self.data_file is not None:
                # Using price stream data file
                
                custom_data_filepath = self.data_file
                
                # Make copy of file to prevent read-write errors
                copy2(custom_data_filepath, self.abs_streamfile_copy)
                
                if int(self.verbosity) > 1:
                    print("Using data file specified ({}).".format(custom_data_filepath))
                
                # Read datafile to get base interval data
                data = pd.read_csv(self.abs_streamfile_copy, 
                                   index_col = 0,
                                   skipinitialspace=True)
                
                if len(data) > 0:
                    data.index = pd.to_datetime(data.index, 
                                                infer_datetime_format=True,
                                                errors='ignore')
                    
                    # Remove copied file
                    os.remove(self.abs_streamfile_copy)
                    
                else:
                    # Stream has not had enough time to write data yet, revert 
                    # to downloading M1 data
                    data = getattr(self.get_data, feed.lower())(instrument,
                                                                granularity = 'M1',
                                                                count = period)
                    
                    if self.check_data_alignment:
                        data = self._verify_data_alignment(data, instrument, feed, period, 
                                                           price_data_path)
                
                # Fetch MTF data
                MTF_data = {self.base_interval: data}
                
                if self.MTF_initialisation:
                    # Download MTF data for strategy initialisation only
                    if self.MTF_data is None:
                        # MTF_data has not been retrieved yet
                        for granularity in self.MTF_intervals:
                            data = getattr(self.get_data, feed.lower())(instrument,
                                                                        granularity = granularity,
                                                                        count=period)
                            
                            if self.check_data_alignment:
                                data = self._verify_data_alignment(data, instrument, feed, period, 
                                                                   price_data_path)
                            
                            MTF_data[granularity] = data
                        
                        self.MTF_data = MTF_data
                        
                    else:
                        # MTF_data already exists, reuse
                        MTF_data = self.MTF_data
                        
                        # Replace base_interval data with latest data
                        MTF_data[self.base_interval] = data
                        
                else:
                    # Download MTF data each update
                    for granularity in self.MTF_intervals:
                        data = getattr(self.get_data, feed.lower())(instrument,
                                                                    granularity = granularity,
                                                                    count=period)
                        
                        if self.check_data_alignment:
                            data = self._verify_data_alignment(data, instrument, feed, period, 
                                                               price_data_path)
                        
                        MTF_data[granularity] = data
                        
                        
                if len(MTF_data) == 1:
                    # There is only one timeframe of data
                    MTF_data = None
                    
            else:            
                # Running in periodic-download mode
                MTF_data = {}
                for granularity in interval.split(','):
                    if granularity == "tick":
                        print("Warning: cannot download historic tick data. " + \
                              "Please change candlestick granularity in " + \
                              "strategy configuration. Exiting.")
                        sys.exit(0)
                        
                    data = getattr(self.get_data, feed.lower())(instrument,
                                                                granularity = granularity,
                                                                count=period)
                    
                    if self.check_data_alignment:
                        data = self._verify_data_alignment(data, instrument, feed, period, 
                                                           price_data_path)
                    
                    MTF_data[granularity] = data
                
                first_granularity = interval.split(',')[0]
                data = MTF_data[first_granularity]
                
                if len(MTF_data) == 1:
                        MTF_data = None
            
            return data, None, MTF_data


    def _verify_data_alignment(self, data, instrument, feed, period, price_data_path):
        '''
        Verifies data time-alignment based on current time and last
        candle in data.
        '''
        interval = self.strategy_params['granularity']
        
        # Check data time alignment
        current_time        = datetime.now(tz=pytz.utc)
        last_candle_closed  = self.broker_utils.last_period(current_time, interval)
        data_ts             = data.index[-1].to_pydatetime().timestamp()
        
        if data_ts != last_candle_closed.timestamp():
            # Time misalignment detected - attempt to correct
            count = 0
            while data_ts != last_candle_closed.timestamp():
                print("  Time misalginment detected at {}".format(datetime.now().strftime("%H:%M:%S")),
                      "({}/{}).".format(data.index[-1].minute, last_candle_closed.minute),
                      "Trying again...")
                time.sleep(3) # wait 3 seconds...
                data    = getattr(self.get_data, feed.lower())(instrument,
                                    granularity = interval,
                                    count=period)
                data_ts = data.index[-1].to_pydatetime().timestamp()
                count   += 1
                if count == 3:
                    break
            
            if data_ts != last_candle_closed.timestamp():
                # Time misalignment still present - attempt to correct
                # Check price data directory to see if the stream has caught 
                # the latest candle
                price_data_filename = "{0}{1}.txt".format(interval, instrument)
                abs_price_path      = os.path.join(price_data_path, price_data_filename)
                
                if os.path.exists(abs_price_path):
                    # Price data file matching instrument and granularity 
                    # exists, check latest candle in file
                    f                   = open(abs_price_path, "r")
                    price_lines         = f.readlines()
                    
                    if len(price_lines) > 1:
                        latest_candle       = price_lines[-1].split(',')
                        latest_candle_time  = datetime.strptime(latest_candle[0],
                                                                '%Y-%m-%d %H:%M:%S')
                        UTC_last_candle_in_file = latest_candle_time.replace(tzinfo=pytz.UTC)
                        price_data_ts       = UTC_last_candle_in_file.timestamp()
                        
                        if price_data_ts == last_candle_closed.timestamp():
                            data    = self.broker_utils.update_data_with_candle(data, latest_candle)
                            data_ts = data.index[-1].to_pydatetime().timestamp()
                            print("  Data updated using price stream.")
            
            # if data is still misaligned, perform manual adjustment.
            if data_ts != last_candle_closed.timestamp():
                print("  Could not retrieve updated data. Aborting.")
                sys.exit(0)
        
        return data
    
    
    def _update(self, i):
        '''
        Update strategy with latest data and generate latest signal.
        '''
        
        # First clear self.latest_orders
        self.latest_orders = []
        
        if self.scan_mode:
            open_positions      = None
        else:
            open_positions      = self.broker.get_open_positions(self.instrument)
        
        # Run strategy to get signals
        signal_dict = self.strategy.generate_signal(i, open_positions)
        
        if 0 not in signal_dict:
            # Single order signal, nest in dictionary to allow iteration
            signal_dict = {1: signal_dict}
            
        # Begin iteration over signal_dict to extract each order
        for order in signal_dict:
            order_signal_dict = signal_dict[order].copy()
            
            if order_signal_dict["direction"] != 0:
                self._process_signal(order_signal_dict, i, self.data, 
                                    self.quote_data, self.instrument)
        
        # TODO - implement the following, maybe
        # else:
        #     signal_type = order_signal_dict["signal_type"] if "signal_type" in order_signal_dict else None
            
        #     if signal_type == 'deployment':
        #         # Strategy deployment signal
        #         runfile = os.path.join(self.home_dir, order_signal_dict['runfile'])
        #         os.system("nohup python3 {} &".format(runfile))
        
        if int(self.verbosity) > 1:
            if len(self.latest_orders) > 0:
                for order in self.latest_orders:
                    order_string = "{}: {} {}".format(order['order_time'], 
                                                      order['instrument'], 
                                                      order['order_type']) + \
                        " order of {} units placed at {}.".format(order['size'],
                                                                  order['order_price'])
                    print(order_string)
            else:
                print("{}: No signal detected ({}).".format(datetime.now().strftime("%b %d %H:%M:%S"),
                                                            self.instrument))
        
        # Check for orders placed and/or scan hits
        if int(self.notify) > 0 and self.backtest_mode is False:
            
            for order_details in self.latest_orders:
                self.broker_utils.write_to_order_summary(order_details, 
                                                         self.order_summary_fp)
            
            if int(self.notify) > 1 and \
                self.email_params['mailing_list'] is not None and \
                self.email_params['host_email'] is not None:
                    if int(self.verbosity) > 0 and len(self.latest_orders) > 0:
                            print("Sending emails ...")
                            
                    for order_details in self.latest_orders:
                        emailing.send_order(order_details,
                                            self.email_params['mailing_list'],
                                            self.email_params['host_email'])
                        
                    if int(self.verbosity) > 0 and len(self.latest_orders) > 0:
                            print("  Done.\n")
            
        # Check scan results
        if self.scan_mode:
            # Construct scan details dict
            scan_details    = {'index'      : self.scan_index,
                               'strategy'   : self.strategy.name,
                               'timeframe'  : self.strategy_params['granularity']
                                }
            
            # Report AutoScan results
            # Scan reporting with no emailing requested.
            if int(self.verbosity) > 0 or \
                int(self.notify) == 0:
                if len(self.scan_results) == 0:
                    print("{}: No signal detected.".format(self.instrument))
                else:
                    # Scan detected hits
                    for instrument in self.scan_results:
                        signal = self.scan_results[instrument]['signal']
                        signal_type = 'Long' if signal == 1 else 'Short'
                        print(f"{instrument}: {signal_type} signal detected.")
            
            if int(self.notify) > 0:
                # Emailing requested
                if len(self.scan_results) > 0 and \
                    self.email_params['mailing_list'] is not None and \
                    self.email_params['host_email'] is not None:
                    # There was a scanner hit and email information is provided
                    emailing.send_scan_results(self.scan_results, 
                                               scan_details, 
                                               self.email_params['mailing_list'],
                                               self.email_params['host_email'])
                elif int(self.notify) > 1 and \
                    self.email_params['mailing_list'] is not None and \
                    self.email_params['host_email'] is not None:
                    # There was no scan hit, but notify set > 1, so send email
                    # regardless.
                    emailing.send_scan_results(self.scan_results, 
                                               scan_details, 
                                               self.email_params['mailing_list'],
                                               self.email_params['host_email'])
                    
    
    def _update_backtest(self, i):
        '''
        Updates virtual broker with latest price data.
        '''
        candle = self.data.iloc[i]
        self.broker.update_positions(candle, self.instrument)
    
    
    def _process_signal(self, order_signal_dict, i, data, quote_data, 
                       instrument):
        '''
            Process order_signal_dict and send orders to broker.
        '''
        signal = order_signal_dict["direction"]
        
        # Entry signal detected, get price data
        price_data      = self.broker.get_price(instrument=instrument, 
                                                data=data, 
                                                conversion_data=quote_data, 
                                                i=i)
        datetime_stamp  = data.index[i]
        
        if signal < 0:
            order_price = price_data['bid']
            HCF         = price_data['negativeHCF']
        else:
            order_price = price_data['ask']
            HCF         = price_data['positiveHCF']
        
        
        # Define 'working_price' to calculate size and TP
        if order_signal_dict["order_type"] == 'limit' or order_signal_dict["order_type"] == 'stop-limit':
            working_price = order_signal_dict["order_limit_price"]
        else:
            working_price = order_price
        
        # Calculate exit levels
        pip_value   = self.broker_utils.get_pip_ratio(instrument)
        stop_distance = order_signal_dict['stop_distance'] if 'stop_distance' in order_signal_dict else None
        stop_type = order_signal_dict['stop_type'] if 'stop_type' in order_signal_dict else None
        
        if 'stop_loss' not in order_signal_dict and \
            'stop_distance' in order_signal_dict and \
            order_signal_dict['stop_distance'] is not None:
            stop_price = working_price - np.sign(signal)*stop_distance*pip_value
        else:
            stop_price = order_signal_dict['stop_loss'] if 'stop_loss' in order_signal_dict else None
        
        if 'take_profit' not in order_signal_dict and \
            'take_distance' in order_signal_dict and \
            order_signal_dict['take_distance'] is not None:
            # Take profit distance specified
            take_profit = working_price + np.sign(signal)*order_signal_dict['take_distance']*pip_value
        else:
            # Take profit price specified, or no take profit specified at all
            take_profit = order_signal_dict["take_profit"] if 'take_profit' in order_signal_dict else None
        
        # Calculate risked amount
        amount_risked = self.broker.get_balance() * self.strategy_params['risk_pc'] / 100
        
        # Calculate size
        if 'size' in order_signal_dict:
            size = order_signal_dict['size']
        else:
            if self.strategy_params['sizing'] == 'risk':
                size            = self.broker_utils.get_size(instrument,
                                                 amount_risked, 
                                                 working_price, 
                                                 stop_price, 
                                                 HCF,
                                                 stop_distance)
            else:
                size = self.strategy_params['sizing']
        
        # Construct order dict by building on signal_dict
        order_details                   = order_signal_dict
        order_details["order_time"]     = datetime_stamp
        order_details["strategy"]       = self.strategy.name
        order_details["instrument"]     = instrument
        order_details["size"]           = signal*size
        order_details["order_price"]    = order_price
        order_details["HCF"]            = HCF
        order_details["granularity"]    = self.strategy_params['granularity']
        order_details["stop_distance"]  = stop_distance
        order_details["stop_loss"]      = stop_price
        order_details["take_profit"]    = take_profit
        order_details["stop_type"]      = stop_type
        order_details["related_orders"] = order_signal_dict['related_orders'] if 'related_orders' in order_signal_dict else None

        # Place order
        if self.scan_mode:
            # Bot is scanning
            scan_hit = {"size"  : size,
                        "entry" : order_price,
                        "stop"  : stop_price,
                        "take"  : take_profit,
                        "signal": signal
                        }
            self.scan_results[instrument] = scan_hit
            
        else:
            # Bot is trading
            self.broker.place_order(order_details)
            self.latest_orders.append(order_details)
    
    def _next_candle_open(self, granularity):
        '''
        Returns the UTC datetime object corresponding to the open time of the 
        next candle.
        '''
        
        current_ts = datetime.now(tz=pytz.utc).timestamp()
        granularity_in_seconds = self.broker_utils.interval_to_seconds(granularity)
        next_candle_open_ts = granularity_in_seconds * np.ceil(current_ts / granularity_in_seconds)
        
        return datetime.fromtimestamp(next_candle_open_ts, tz=pytz.utc)
            

    def create_backtest_summary(self, balance, NAV, margin):
        '''
        Constructs backtest summary dictionary for further processing.
        '''
        
        trade_summary = self.broker_utils.trade_summary(self.instrument, self.broker.closed_positions)
        open_trade_summary = self.broker_utils.open_order_summary(self.instrument, self.broker.open_positions)
        cancelled_summary = self.broker_utils.cancelled_order_summary(self.instrument, self.broker.cancelled_orders)
            
        backtest_dict = {}
        backtest_dict['data']           = self.data
        backtest_dict['NAV']            = NAV
        backtest_dict['balance']        = balance
        backtest_dict['margin']         = margin
        backtest_dict['trade_summary']  = trade_summary
        backtest_dict['indicators']     = self.strategy.indicators if hasattr(self.strategy, 'indicators') else None
        backtest_dict['instrument']     = self.instrument
        backtest_dict['interval']       = self.strategy_params['granularity']
        backtest_dict['open_trades']    = open_trade_summary
        backtest_dict['cancelled_trades'] = cancelled_summary
        
        self.backtest_summary = backtest_dict
    
    def _get_iteration_range(self):
        '''
        Checks mode of operation and returns data iteration range. For backtesting,
        the entire dataset is iterated over. For livetrading, only the latest candle
        is used.
        '''
        
        if self.backtest_mode:
            start_range = 0
        else:
            start_range = len(self.data)-1
        end_range       = len(self.data)

        return start_range, end_range
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
AutoStream
-----------
Data stream function for Oanda v20 API.
"""

import v20
import json
from datetime import datetime
import sys
import calendar
import time
import re
import os
import traceback
import pandas as pd


class stream_record(object):
    ''' 
    Creates a stream record.
    
    Attributes
    ----------
        data : dict
            A dictionary containing information of the tick; instrument, time,
            bid, ask and mid.
        
        record_type : str
            The type of stream record: either PRICE or HEARTBEAT
    '''
    
    def __init__(self, msg):
        self.data           = {}
        self.record_type    = msg['type']
        
        if msg['type'] == 'PRICE':
            self.dt                  = datetime.strptime(msg['time'][:-4], '%Y-%m-%dT%H:%M:%S.%f')
            self.epoch               = int(calendar.timegm(self.dt.timetuple()))
            self.data['instrument']  = msg["instrument"]
            self.data['time']        = msg["time"]
            self.data['bid']         = msg["closeoutBid"]
            self.data['ask']         = msg["closeoutAsk"]
            self.data['mid']         = (float(msg['closeoutBid']) + float(msg['closeoutAsk']))/2.0
            
        elif msg['type'] == "HEARTBEAT":
            self.dt                  = datetime.strptime(msg['time'][:-4], '%Y-%m-%dT%H:%M:%S.%f')
            self.epoch               = int(calendar.timegm(self.dt.timetuple()))
            self.data['time']        = msg["time"]


class candle_builder(object):
    ''' Builds candles from stream records '''
    
    def __init__(self, instrument, granularity):
        ''' Initialises class based on instrument and candlestick 
            granularity. 
        '''
        self.instrument     = instrument
        self.duration       = self.granularity_to_seconds(granularity)
        self.granularity    = granularity
        self.data           = None
        self.start          = None
        self.end            = None
        
    def initialise_data(self, tick):
        ''' Cosntructs data dictionary for new candle. '''
        self.start  = tick.epoch - (tick.epoch % self.duration)
        self.end    = tick.epoch - (tick.epoch % self.duration) + self.duration
        self.data   = {"instrument"     : self.instrument,
                       "start"          : "%s" % self.seconds_in_time(self.start),
                       "end"            : "%s" % self.seconds_in_time(self.end),
                       "granularity"    : self.granularity,
                       "completed"      : False,
                       "data"           : {"open"   : tick.data['mid'],
                                           "high"   : tick.data['mid'],
                                           "low"    : tick.data['mid'],
                                           "last"   : tick.data['mid'],
                                           "volume" : 1
                                            }
                       }

    def seconds_in_time(self, e):
        ''' Converts a timestamp to datetime object. '''
        w   = time.gmtime(e)
        
        return datetime(*list(w)[0:6])

    def make_candle(self, completed=False):
        ''' Closes candle '''
        
        self.data['completed'] = completed
        
        return self.data.copy()
    
    def granularity_to_seconds(self, granularity):
        ''' Converts a granularity to time in seconds '''
        mfact = {'S': 1,
                 'M': 60,
                 'H': 3600,
                 'D': 86400,
                 }
    
        f, n    = re.match("(?P<f>[SMHD])(?:(?P<n>\d+)|)", granularity).groups()
        n       = n if n else 1
        seconds = mfact[f] * int(n)
        
        return seconds
    
    def process_tick(self, tick):
        ''' Processes tick into candle '''
        if tick.record_type == 'HEARTBEAT':
            if self.data and tick.epoch > self.end:
                # this frame is completed based on the heartbeat timestamp
                candle      = self.make_candle(completed=True)
                self.data   = None     # clear it, reinitialized by the next tick
                
                print("infrequent ticks: %s, %s completed with "
                            "heartbeat (%d secs)" %
                            (self.instrument, self.granularity,
                              (tick.epoch - self.end)))
                
                return candle
            
            else:
                # Candle is still open
                return
            
        
        if not tick.data['instrument'] == self.instrument:
            # Tick data is not for instrument being monitored - exit
            
            return
        
        
        if not self.data:
            # There is no data yet, initialise now
            self.initialise_data(tick)
            
            return None
        
        
        if tick.epoch >= self.start and tick.epoch < self.end:
            # The tick falls within the current candle
            
            if tick.data['mid'] > self.data['data']['high']:
                self.data['data']['high'] = tick.data['mid']
                
            if tick.data['mid'] < self.data['data']['low']:
                self.data['data']['low'] = tick.data['mid']
                
            if tick.data['mid'] != self.data['data']['last']:
                self.data['data']['last'] = tick.data['mid']
                
            # Add tick to candle volume
            self.data['data']['volume'] += 1
            
            return None
        
        # This tick is not within the candle, close the candle.
        candle = self.make_candle(completed = True)
        
        self.initialise_data(tick)
        
        return candle


class AutoStream():
    '''
    AutoStream Class.
    ------------------
    
    Methods:
        main(): Subscribes to stream and builds candlestick price files. 
    
    
    Attributes:
        home_dir : str
            The path of the home directory. If writing stream to file, it will
            be written to home_dir/price_data.
        
        stream_config : dict
            Dictionary containing stream configuration information.
        
        instruments : list
            The instruments to be streamed.
        
        granularity : str
            The granularity of candlesticks to build from the stream.
        
        no_candles : int
            The maximum number of candles to write to file.
        
        record_ticks : bool
            Flag to capture tick data.
        
        record_candles : bool
            Flag to capture candlestick data.
        
        write_to_file : bool
            Flag to write streamed data to csv files.
    '''
    
    def __init__(self, home_dir, stream_config, 
                 instrument, granularity=None, no_candles=10,
                 record_ticks=False, record_candles=False,
                 bot=None, update_bot=False, write_to_file=False):
        '''
        Assign attributes required to stream.
        '''
        
        self.home_dir       = home_dir
        self.instruments    = instrument
        self.granularity    = granularity
        self.no_candles     = no_candles
        self.record_candles = record_candles
        self.record_ticks   = record_ticks
        self.bot            = bot
        self.update_bot     = update_bot
        self.write_to_file  = write_to_file
        self.killfile       = os.path.join(home_dir, 'stopstream')
        self.suspendfile    = os.path.join(home_dir, 'suspendstream')
        self.stopstream     = False
        
        # Runtime attributes
        self.tick_data      = None
        self.candle_data    = None
        
        # Add instruments to stream_config
        stream_config['instruments'] = instrument
        self.stream_config  = stream_config
        
        # Perform initialisation checks
        self.checks_passed = True
        if not record_candles and not record_ticks:
            print("Please specify whether to record ticks, candles or both "+\
                  "using the record_ticks and record_candles attributes.")
            self.checks_passed = False
        
        if record_candles and granularity is None:
            print("Please provide a candlestick granularity when streaming candles.")
            self.checks_passed = False
            
    
    def start(self):
        '''
        Starts stream if all checks have passed.
        '''

        if self.checks_passed:
            # Proceed to run stream
            self.main()
        
    def main(self):
        '''
        Subscribes to stream and builds candlestick price files. 
        '''
        
        data_dir_path   = os.path.join(self.home_dir, 'price_data')
        
        # Initialise tick DataFrame
        if self.record_ticks:
            self.tick_data = pd.DataFrame()
        
        # Initialise candle DataFrame
        if self.record_candles:
            self.candle_data = pd.DataFrame()
        
        # Initialise text file processing
        candle_filenames = {}
        tick_filenames = {}
        candle_builders = {}
        
        if self.write_to_file:
            if not os.path.exists(data_dir_path):
                # The price_data directory doesn't exist, make it
                os.makedirs(data_dir_path)
        
            # Initialise candle factories
            for instrument in self.instruments.split(','):
                if self.record_candles:
                    filename                    = "{0}{1}.txt".format(self.granularity, 
                                                                      instrument)
                    abs_filename                = os.path.join(data_dir_path, filename)
                    candle_filenames[instrument] = abs_filename
                    candle_builders[instrument] = candle_builder(instrument, 
                                                                 self.granularity)
                
                if self.record_ticks:
                    filename                    = "{}_ticks.txt".format(instrument)
                    abs_filename                = os.path.join(data_dir_path, filename)
                    tick_filenames[instrument]  = abs_filename
        
        # Connect to stream and begin processing 
        self.connect_to_stream(self.stream_config)
        
        while True and not self.stopstream:
            try:
                self.process_stream(candle_builders,
                                    candle_filenames,
                                    tick_filenames)
            except BaseException as ex:
                ex_type, ex_value, ex_traceback = sys.exc_info()
                        
                # Extract unformatter stack traces as tuples
                trace_back = traceback.extract_tb(ex_traceback)
            
                # Format stacktrace
                stack_trace = list()
            
                for trace in trace_back:
                    trade_string = "File : %s , Line : %d, " % (trace[0], trace[1]) + \
                                   "Func.Name : %s, Message : %s" % (trace[2], trace[3])
                    stack_trace.append(trade_string)
                
                print("\nWARNING FROM AUTOSTREAM ({}): The following exception was caught:".format(self.instruments))
                print("Time: {}".format(datetime.now().strftime("%b %d %H:%M:%S")))
                print("Exception type : %s " % ex_type.__name__)
                print("Exception message : %s" %ex_value)
                print("Stack trace : %s" %stack_trace)
                print("  Trying again.")
                            
                # Sleep and attempt to re-connect to stream
                time.sleep(3)
                self.connect_to_stream(self.stream_config)
    
    
    def connect_to_stream(self, config):
        ''' Connects to Oanda streaming API '''
        ACCESS_TOKEN    = config["ACCESS_TOKEN"]
        port            = config["PORT"]
        ACCOUNT_ID      = config["ACCOUNT_ID"]
        STREAM_API      = "stream-fxpractice.oanda.com"
        instruments     = config["instruments"]
        
        streamAPI       = v20.Context(hostname = STREAM_API,
                                      token = ACCESS_TOKEN, 
                                      port = port)
        
        # Connect to the stream
        for attempt in range(5):
            try:
                response = streamAPI.pricing.stream(accountID = ACCOUNT_ID, 
                                                    instruments = instruments,
                                                    snapshot = True
                                                    )
                if response.status != 200:
                    print("Warning:")
                    print("Time: {}".format(datetime.now().strftime("%b %d %H:%M:%S")))
                    print(response.reason)
                    break
                    
                else:
                    self.stream = response
            
            except Exception as e:
                print("Caught exception when connecting to stream\n" + str(e))
                time.sleep(3)
            else:
                break
            
        else:
            print("All attempts to connect to stream failed. Exiting.")
            sys.exit(0)
    
    def suspend(self):
        while os.path.exists(self.suspendfile):
            pass
        
    def process_stream(self, candle_builders, candle_filenames, tick_filenames):
        '''
        Processes stream based on run settings.
        '''
        
        print("\nProcessing stream. To stop streaming, create file named " +\
              "stopstream in the home directory. Alternatively, create " +\
              "a file called suspendstream to pause the stream. Then, " +\
              "delete this file to resume streaming.")
        print("Home directory: ", self.home_dir)

        for line in self.stream.lines:
            
            # First check for stop file
            if os.path.exists(self.killfile):
                print("Stop file deteced. Stream stopping.")
                self.stopstream = True
                
                break
            
            # Next check for suspend file
            if os.path.exists(self.suspendfile):
                print("Suspending stream.")
                self.suspend()
                print("Resuming stream.")
            
            # Process each update of stream
            line    = line.decode('utf-8')
            msg     = json.loads(line)
            tick    = stream_record(msg)
            
            if self.record_ticks and msg['type'] == 'PRICE' and tick.data['instrument'] == self.instruments:
                # Create tick df from stream record
                new_tick = {'Bid': tick.data['bid'], 
                            'Ask': tick.data['ask'], 
                            'Mid': tick.data['mid']}
                latest_tick = pd.DataFrame(new_tick, 
                                           index=[tick.data['time']])
                
                # Update tick_data with latest tick
                self.tick_data = self.tick_data.append(latest_tick)
                
                # Check if the length has been exceeded
                if len(self.tick_data) > self.no_candles:
                    self.tick_data = self.tick_data.iloc[-self.no_candles:, :]
                
                # Update bot
                if self.update_bot:
                    self.update_bot_data(self.tick_data)
                
                # Write to file
                if self.write_to_file:
                    # TODO - preprocess time index column to convert to datetime
                    # And add exception handling methods for datetime errors 
                    self.tick_data.index.name = 'Time'
                    self.tick_data.to_csv(tick_filenames[tick.data['instrument']])
                
            if self.record_candles:
                for instrument in candle_builders:
                    
                    candle  = candle_builders[instrument].process_tick(tick)
                    
                    if candle is not None:
                        
                        # Create candle df from stream record
                        new_candle = {'High': candle['data']['high'], 
                                      'Low': candle['data']['low'], 
                                      'Open': candle['data']['open'],
                                      'Close': candle['data']['last']}
                        latest_candle = pd.DataFrame(new_candle, 
                                                     index=[candle['start']])
                        
                        # Update candle_data with latest candle
                        self.candle_data = self.candle_data.append(latest_candle)
                        
                        # Check if the length has been exceeded
                        if len(self.candle_data) > self.no_candles:
                            self.candle_data = self.candle_data.iloc[-self.no_candles:, :]
                        
                        # Update bot
                        if self.update_bot:
                            self.update_bot_data(self.candle_data)
                        
                        # Write to file
                        if self.write_to_file:
                            self.candle_data.index.name = 'Time'
                            self.candle_data.to_csv(candle_filenames[instrument])
            

    def update_bot_data(self, data):
        '''
        Sends updated data to bot.
        '''
        
        # Only pass data if len(data) > 0
        if len(data) > 0:
            # Refresh strategy with latest data
            self.bot._update_strategy_data(data)
            
            # Call bot update to act on latest data
            self.bot._update(-1)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
---------------------------------------------------------------------------
         _   _   _ _____ ___ _____ ____      _    ____  _____ ____  
        / \ | | | |_   _/ _ \_   _|  _ \    / \  |  _ \| ____|  _ \ 
       / _ \| | | | | || | | || | | |_) |  / _ \ | | | |  _| | |_) |
      / ___ \ |_| | | || |_| || | |  _ <  / ___ \| |_| | |___|  _ < 
     /_/   \_\___/  |_| \___/ |_| |_| \_\/_/   \_\____/|_____|_| \_\
    
---------------------------------------------------------------------------
     A Python-Based Development Platform For Automated Trading Systems
"""

from datetime import datetime, timedelta
import sys
import os
import pyfiglet
import importlib
import numpy as np
import pandas as pd
import timeit
from scipy.optimize import brute
from ast import literal_eval
from autotrader.brokers.oanda import Oanda
from autotrader.brokers.virtual.virtual_broker import Broker
from autotrader.lib import instrument_list, environment_manager, printout
from autotrader.lib.read_yaml import read_yaml
from autotrader import autoplot
from autotrader.autobot import AutoTraderBot
from autotrader.lib.bot_manager import ManageBot
import time

class AutoTrader():
    """
    AutoTrader: A Python-Based Development Platform For Automated Trading Systems.
    ------------------------------------------------------------------------------
    Website: https://kieran-mackle.github.io/AutoTrader/
    
    GitHub: https://github.com/kieran-mackle/AutoTrader
    
    Author: Kieran Mackle
    
    Version: 0.4.x

    Attributes
    ----------
    Note: many of the following attributes are set from the configure method of AutoTrader.
    
    feed : str 
        The data feed to be used (eg. Yahoo, Oanda).
                
    verbosity : int
        The verbosity of AutoTrader (0, 1 or 2).
    
    notify : int
        The level of email notification (0, 1 or 2).
    
    home_dir : str 
        The project home directory.
    
    use_stream : bool 
        Set to True to use price stream as data feed.
    
    detach_bot : bool 
        Set to True to spawn new thread for each bot deployed. Bots will then
        continue to trade until a termination signal is recieved from the strategy.
    
    check_data_alignment : bool
        Verify time of latest candle in data recieved against current time.
    
    allow_dancing_bears : bool
        Allow incomplete candles to be passed to strategy.
    
    account_id : str
        The brokerage account ID to use in this instance.
    
    environment : str
        The trading environment of this instance.
    
    show_plot : bool
        Automatically display plot of results.
    
    MTF_initialisation : bool
        Only download mutliple time frame data when initialising the strategy, 
        rather than every update.


    Methods
    -------
    run():
        Runs AutoTrader.
    
    configure(feed='yahoo', verbosity=1, notify=0, home_dir=None,
              use_stream=False, detach_bot=False,
              check_data_alignment=True, allow_dancing_bears=False,
              account_id=None, environment='demo', show_plot=False,
              MTF_initialisation=False):
        Configures various run settings for AutoTrader.
    
    add_strategy(strategy_filename=None, strategy_dict=None)
        Adds a strategy to AutoTrader. 
    
    plot_backtest(bot=None):
        Plots backtest results of an AutoTrader Bot.
    
    """
    
    def __init__(self):
        '''
        AutoTrader initialisation. Called when creating new AutoTrader instance.
        '''
        
        self.home_dir       = None
        self.order_summary_fp = None
        
        self.verbosity      = 1
        self.notify         = 0
        self.email_params   = None
        self.show_help      = None
        self.show_plot      = False
        
        # Livetrade Parameters
        self.detach_bot     = False
        self.check_data_alignment = True
        self.allow_dancing_bears = False
        self.use_stream     = False
        self.MTF_initialisation = False
        self.stream_config  = None
        
        self.broker         = None
        self.broker_utils   = None
        self.environment    = 'demo'
        self.account_id     = None
        
        self.strategies     = {}
        self._uninitiated_strat_files = []
        self._uninitiated_strat_dicts = []
        self.feed           = 'yahoo'
        self.bots_deployed  = []
        
        # Backtesting Parameters
        self.backtest_mode = False
        self.data_start = None
        self.data_end   = None
        self.data_file  = None
        self.backtest_initial_balance = None
        self.backtest_spread = None
        self.backtest_commission = None
        self.backtest_leverage = None
        self.backtest_base_currency = None
        
        # Optimisation Parameters
        self.optimisation_config = None
        self.optimise_mode = False
        self.opt_params = None
        self.bounds = None
        self.Ns = None
        
        # Scan Parameters
        self.scan_mode = False
        self.scan_index = None
        self.scan_watchlist = None
        self.scan_results = {}
        
        # Plotting
        self._custom_ap_settings = False
        
    def run(self):
        '''
        Run AutoTrader.
        '''
        
        # Define home_dir if undefined
        if self.home_dir is None:
            self.home_dir = os.getcwd()
        
        # Load uninitiated strategies
        for strat_dict in self._uninitiated_strat_dicts:
            self.add_strategy(strategy_dict=strat_dict)
        for strat_config_file in self._uninitiated_strat_files:
            self.add_strategy(strategy_filename=strat_config_file)
        
        if self.scan_watchlist is not None:
            # Scan watchlist has not overwritten strategy watchlist
            self._update_strategy_watchlist()
        
        # Print help
        if self.show_help is not None:
            printout.option_help(self.show_help)
        
        if len(self.strategies) == 0:
            print("Error: no strategy has been provided. Do so by using the" +\
                  " 'add_strategy' method of AutoTrader.")
            sys.exit(0)
            
        if sum([self.backtest_mode, self.scan_mode]) > 1:
            print("Error: backtest mode and scan mode are both set to True," +\
                  " but only one of these can run at a time.")
            print("Please check your inputs and try again.")
            sys.exit(0)
        
        if self.backtest_mode:
            if self.notify > 0:
                print("Warning: notify set to {} ".format(self.notify) + \
                      "during backtest. Setting to zero to prevent emails.")
                self.notify = 0
        
        if self.optimise_mode:
            if self.backtest_mode:
                self._run_optimise()
            else:
                print("Please set backtest parameters to run optimisation.")
        else:
            self._main()
    
    def usage(self):
        '''
        Prints usage instructions for AutoTrader.
        '''
        printout.usage()
    
    def option_help(self, option):
        '''
        Prints help for a user option of AutoTrader.
        
            Parameters:
                option (str): user option to request help for.
        '''
        printout.option_help(option)
        
    def _main(self):
        '''
        Main run file of autotrader.py. This method is called internally 
        from the "run" method.
        '''
        
        ''' -------------------------------------------------------------- '''
        '''                         Load configuration                     '''
        ''' -------------------------------------------------------------- '''
        # Construct broker config
        global_config_fp = os.path.join(self.home_dir, 'config', 'GLOBAL.yaml')
        if os.path.isfile(global_config_fp):
            global_config = read_yaml(global_config_fp)
        else:
            global_config = None
        broker_config = environment_manager.get_config(self.environment,
                                                       global_config,
                                                       self.feed)
        
        # Construct stream_config dict
        if self.use_stream:
            self.stream_config = broker_config
        
        if self.account_id is not None:
            # Overwrite default account in global config
            broker_config['ACCOUNT_ID'] = self.account_id
        
        self._assign_broker(broker_config)
        self._configure_emailing(global_config)
        
        if self.backtest_mode:
            starting_balance = self.broker.get_balance()
            NAV     = []
            balance = []
            margin  = []
        
        if int(self.verbosity) > 0:
            if self.backtest_mode:
                print("Beginning new backtest.")
                print("  From: ", datetime.strftime(self.data_start,'%d/%m/%Y %H:%M'))
                print("  To:   ", datetime.strftime(self.data_end,'%d/%m/%Y %H:%M'))
                # print("  Instruments: ", self.watchlist)
            elif self.scan_mode:
                print("AutoTrader - AutoScan")
                print("Time: {}\n".format(datetime.now().strftime("%A, %B %d %Y, "+
                                                                  "%H:%M:%S")))
            else:
                print("AutoTrader Livetrade")
                print("--------------------")
                print("Time: {}\n".format(datetime.now().strftime("%A, %B %d %Y, "+
                                                                  "%H:%M:%S")))
        
        ''' -------------------------------------------------------------- '''
        '''    Assign strategy to bot for each instrument in watchlist     '''
        ''' -------------------------------------------------------------- '''
        for strategy in self.strategies:
            for instrument in self.strategies[strategy]['WATCHLIST']:
                bot = AutoTraderBot(instrument, self.strategies[strategy],
                                        self.broker, self)
                
                if self.detach_bot is True and self.backtest_mode is False:
                    # Send bot to bot manager to monitor stream
                    print("Passing bot to bot manager...")
                    bot_name_string = "{}_{}_{}".format(strategy.replace(' ',''),
                                                        self.strategies[strategy]['INTERVAL'].split(',')[0],
                                                        instrument)
                    ManageBot(bot, self.home_dir, bot_name_string, self.use_stream)
                else:
                    self.bots_deployed.append(bot)
                    
        
        ''' -------------------------------------------------------------- '''
        '''                  Analyse price data using strategy             '''
        ''' -------------------------------------------------------------- '''
        if int(self.verbosity) > 0 and self.backtest_mode:
            print("\nTrading...")
        
        # TODO - add check that data ranges are consistent across bots
        # For now, assume correct and use first bot.
        if not self.detach_bot:
            start_range, end_range = self.bots_deployed[0]._get_iteration_range()
            for i in range(start_range, end_range):
                
                # Update each bot with latest data to generate signal
                for bot in self.bots_deployed:
                    bot._update(i)
                    
                    # If backtesting, update virtual broker with latest data
                    if self.backtest_mode:
                        bot._update_backtest(i)
                
                if self.backtest_mode is True:
                    NAV.append(self.broker.NAV)
                    balance.append(self.broker.portfolio_balance)
                    margin.append(self.broker.margin_available)
        
        ''' -------------------------------------------------------------- '''
        '''                     Backtest Post-Processing                   '''
        ''' -------------------------------------------------------------- '''
        # Data iteration complete - proceed to post-processing
        if self.backtest_mode is True:
            # Create backtest summary for each bot 
            for bot in self.bots_deployed:
                bot.create_backtest_summary(balance, NAV, margin)            
            
            if int(self.verbosity) > 0:
                print("\nBacktest complete.")
                if len(self.bots_deployed) == 1:
                    bot = self.bots_deployed[0]
                    trade_summary = bot.backtest_summary['trade_summary']
                    backtest_results = self.extract_backtest_results(trade_summary, 
                             self.broker, starting_balance, self.broker_utils) 
                    self.print_backtest_results(backtest_results)
                    
                else:
                    self.multibot_backtest_results = self.multibot_backtest_analysis()
                    self.print_multibot_backtest_results(self.multibot_backtest_results)
                    
                    print("Results for multiple-instrument backtests have been")
                    print("written to AutoTrader.multibot_backtest_results.")
                    print("Individual bot results can be found in AutoTrader.bots_deployed.")
            
            if self.show_plot:
                if len(self.bots_deployed) == 1:
                    if len(self.bots_deployed[0].backtest_summary['trade_summary']) > 0:
                        self.plot_backtest(bot=self.bots_deployed[0])
                
                else:
                    # Backtest run with multiple bots
                    cpl_dict = {}
                    for bot in self.bots_deployed:
                        
                        profit_df = pd.merge(bot.data, 
                                 bot.backtest_summary['trade_summary']['Profit'], 
                                 left_index=True, right_index=True).Profit.cumsum()
                        cpl_dict[bot.instrument] = profit_df
                    
                    ap = autoplot.AutoPlot(bot.data)
                    ap._plot_multibot_backtest(self.multibot_backtest_results, 
                                              NAV,
                                              cpl_dict)
        
        elif self.scan_mode and self.show_plot:
            # Show plots for scanned instruments
            for bot in self.bots_deployed:
                ap = self._instantiate_autoplot(bot.data)
                ap.plot(indicators = bot.strategy.indicators, 
                        instrument = bot.instrument)
                time.sleep(0.3)

    def _clear_strategies(self):
        '''
        Removes all strategies saved in autotrader instance.
        '''
        
        self.strategies = {}
    
    def _clear_bots(self):
        '''
        Removes all deployed bots in autotrader instance.
        '''
        
        self.bots_deployed = []
    
    def plot_settings(self, max_indis_over=3, max_indis_below=2,
                      fig_tools="pan,wheel_zoom,box_zoom,undo,redo,reset,save,crosshair",
                      ohlc_height=400, ohlc_width=800, top_fig_height=150,
                      bottom_fig_height=150):
        ''' Configures settings for AutoPlot. '''
        
        # Set flag for custom settings
        self._custom_ap_settings = True
        
        # Assign attributes
        self.max_indis_over     = max_indis_over
        self.max_indis_below    = max_indis_below
        self.fig_tools          = fig_tools
        self.ohlc_height        = ohlc_height
        self.ohlc_width         = ohlc_width
        self.top_fig_height     = top_fig_height
        self.bottom_fig_height  = bottom_fig_height
    
    def _instantiate_autoplot(self, data):
        ''' Creates instance of AutoPlot. '''
        
        # Create nominal instance
        ap = autoplot.AutoPlot(data)
        
        if self._custom_ap_settings:
            # Assign attributes
            ap.max_indis_over     = self.max_indis_over
            ap.max_indis_below    = self.max_indis_below
            ap.fig_tools          = self.fig_tools
            ap.ohlc_height        = self.ohlc_height
            ap.ohlc_width         = self.ohlc_width
            ap.top_fig_height     = self.top_fig_height
            ap.bottom_fig_height  = self.bottom_fig_height
        
        return ap
    
    def add_strategy(self, strategy_filename=None, 
                     strategy_dict=None):
        '''
        Adds a strategy to AutoTrader. 
        
            Parameters:
                strategy_filename (str): prefix of yaml strategy
                configuration file, located in home_dir/config.
                
                strategy_dict (dict): alternative to strategy_filename,
                the strategy dictionary can be passed directly.
        '''
        
        if self.home_dir is None:
            # Home directory has not yet been set, postpone strategy addition
            if strategy_filename is None:
                self._uninitiated_strat_dicts.append(strategy_dict)
            else:
                self._uninitiated_strat_files.append(strategy_filename)
            
        else:
            if strategy_dict is None:
                config_file_path = os.path.join(self.home_dir, 'config', strategy_filename)
                new_strategy = read_yaml(config_file_path + '.yaml')
            else:
                new_strategy = strategy_dict
            
            name = new_strategy['NAME']
            
            if name in self.strategies:
                print("Warning: duplicate strategy name deteced. Please check " + \
                      "the NAME field of your strategy configuration file and " + \
                      "make sure it is not the same as other strategies being " + \
                      "run from this instance.")
                print("Conflicting name:", name)
            
            self.strategies[name] = new_strategy
    
    
    def backtest(self, start=None, end=None, initial_balance=1000, spread=0, 
                 commission=0, leverage=1, base_currency='AUD', start_dt=None, 
                 end_dt=None):
        '''
        Configures settings for backtesting.
        
            Parameters:
                start (str): start date for backtesting, in format d/m/yyyy.
                
                end (str): end date for backtesting, in format d/m/yyyy.
                
                initial_balance (float): initial account balance in base currency 
                units.
                
                spread (float): bid/ask spread of instrument.
                
                commission (float): trading commission as percentage per trade.
                
                leverage (int): account leverage.
                
                base_currency (str): base currency of account.
                
                start_dt (datetime): datetime object corresponding to start time.
                
                end_dt (datetime): datetime object corresponding to end time.
                
            Note: 
                Start and end times must be specified as the same type. For
                example, both start and end arguments must be provided together, 
                or alternatively, start_dt and end_dt must both be provided.
        '''
        
        # Convert start and end strings to datetime objects
        if start_dt is None and end_dt is None:
            start_dt    = datetime.strptime(start + '+0000', '%d/%m/%Y%z')
            end_dt      = datetime.strptime(end + '+0000', '%d/%m/%Y%z')
        
        # Assign attributes
        self.backtest_mode = True
        self.data_start = start_dt
        self.data_end   = end_dt
        self.backtest_initial_balance = initial_balance
        self.backtest_spread = spread
        self.backtest_commission = commission
        self.backtest_leverage = leverage
        self.backtest_base_currency = base_currency
    
    
    def configure(self, feed='yahoo', verbosity=1, notify=0, home_dir=None,
                  use_stream=False, detach_bot=False,
                  check_data_alignment=True, allow_dancing_bears=False,
                  account_id=None, environment='demo', show_plot=False,
                  MTF_initialisation=False):
        '''
        AutoTrader Run Configuration
        -------------------------------
        
        Configures various run settings for AutoTrader.
        
            Parameters:
                feed (str): the data feed to be used (eg. Yahoo, Oanda).
                
                verbosity (int): the verbosity of AutoTrader (0, 1 or 2).
                
                notify (int): the level of email notification (0, 1 or 2).
                
                home_dir (str): the project home directory.
                
                use_stream (bool): set to True to use price stream as data feed.
                
                detach_bot (bool): set to True to spawn new thread for each bot
                deployed.
                
                check_data_alignment (bool): verify time of latest candle in
                data recieved against current time.
                
                allow_dancing_bears (bool): allow incomplete candles to be 
                passed to strategy.
                
                account_id (str): the brokerage account ID to use in this instance.
                
                environment (str): the trading environment of this instance.
                
                show_plot (bool): automatically display plot of results.
                
                MTF_initialisation (bool): only download mutliple time frame 
                data when initialising the strategy, rather than every update.
        '''
        
        self.feed = feed
        self.verbosity = verbosity
        self.notify = notify
        self.home_dir = home_dir if home_dir is not None else os.getcwd()
        self.use_stream = use_stream
        self.detach_bot = detach_bot
        self.check_data_alignment = check_data_alignment
        self.allow_dancing_bears = allow_dancing_bears
        self.account_id = account_id
        self.environment = environment
        self.show_plot = show_plot
        self.MTF_initialisation = MTF_initialisation
        
    
    def scan(self, strategy_filename=None, strategy_dict=None, scan_index=None):
        '''
        Configure AutoTrader scan. 
            
            Parameters:
                strategy_filename (str): prefix of yaml strategy
                configuration file, located in home_dir/config.
                    
                scan_index (str): index to scan.
        '''
        
        # If a strategy is provided here, add it
        if strategy_filename is not None:
            self.add_strategy(strategy_filename)
        elif strategy_dict is not None:
            self.add_strategy(strategy_dict=strategy_dict)
        
        # If scan index provided, use that. Else, use strategy watchlist
        if scan_index is not None:
            self.scan_watchlist = instrument_list.get_watchlist(scan_index,
                                                                self.feed)

        else:
            scan_index = 'Strategy watchlist'
            
        self.scan_mode = True
        self.scan_index = scan_index
        self.check_data_alignment = False
    
    
    def _update_strategy_watchlist(self):
        ''' Updates the watchlist of each strategy with the scan watchlist. '''
        
        for strategy in self.strategies:
            self.strategies[strategy]['WATCHLIST'] = self.scan_watchlist
    
    
    def plot_backtest(self, bot=None):
        '''
        Plots backtest results of an AutoTrader Bot.
            
            Parameters:
                bot (class): AutoTrader bot class containing backtest results.
        '''
        ap = autoplot.AutoPlot(bot.data)
        profit_df = pd.merge(bot.data, 
                             bot.backtest_summary['trade_summary']['Profit'], 
                             left_index=True, right_index=True).Profit.cumsum()
        
        ap.plot(bot.backtest_summary, cumulative_PL=profit_df)
                
    
    def multibot_backtest_analysis(self, bots=None):
        '''
        Analyses backtest results of multiple bots to create an overall 
        performance summary.
        
            Parameters:
                bots (list): a list of AutoTrader bots to analyse.
        '''
        
        instruments = []
        win_rate    = []
        no_trades   = []
        avg_win     = []
        max_win     = []
        avg_loss    = []
        max_loss    = []
        no_long     = []
        no_short    = []
        
        if bots is None:
            bots = self.bots_deployed
        
        for bot in bots:
            backtest_results = self.analyse_backtest(bot.backtest_summary)
            
            instruments.append(bot.instrument)
            no_trades.append(backtest_results['no_trades'])
            if backtest_results['no_trades'] > 0:
                win_rate.append(backtest_results['all_trades']['win_rate'])
                avg_win.append(backtest_results['all_trades']['avg_win'])
                max_win.append(backtest_results['all_trades']['max_win'])
                avg_loss.append(backtest_results['all_trades']['avg_loss'])
                max_loss.append(backtest_results['all_trades']['max_loss'])
                no_long.append(backtest_results['long_trades']['no_trades'])
                no_short.append(backtest_results['short_trades']['no_trades'])
            else:
                win_rate.append(np.nan)
                avg_win.append(np.nan)
                max_win.append(np.nan)
                avg_loss.append(np.nan)
                max_loss.append(np.nan)
                no_long.append(np.nan)
                no_short.append(np.nan)
            
        
        multibot_backtest_results = pd.DataFrame(data={'win_rate': win_rate,
                                                       'no_trades': no_trades,
                                                       'avg_win': avg_win,
                                                       'max_win': max_win,
                                                       'avg_loss': avg_loss,
                                                       'max_loss': max_loss,
                                                       'no_long': no_long,
                                                       'no_short': no_short},
                                                 index=instruments)
        
        return multibot_backtest_results
        
    
    def analyse_backtest(self, backtest_summary):
        '''
        Analyses backtest summary to extract key statistics.
        
            Parameters:
                backtest_summary (dict): summary of backtest performance of bot.
        '''
        
        trade_summary   = backtest_summary['trade_summary']
        instrument      = backtest_summary['instrument']
        
        cpl             = trade_summary.Profit.cumsum()
        
        backtest_results = {}
        
        # All trades
        no_trades   = len(trade_summary)
        backtest_results['no_trades'] = no_trades
        if no_trades > 0:
            backtest_results['all_trades'] = {}
            wins        = trade_summary[trade_summary.Profit > 0]
            avg_win     = np.mean(wins.Profit)
            max_win     = np.max(wins.Profit)
            loss        = trade_summary[trade_summary.Profit < 0]
            avg_loss    = abs(np.mean(loss.Profit))
            max_loss    = abs(np.min(loss.Profit))
            win_rate    = 100*len(wins)/no_trades
            longest_win_streak, longest_lose_streak  = self.broker_utils.get_streaks(trade_summary)
            avg_trade_duration = np.mean(trade_summary.Trade_duration.values)
            min_trade_duration = min(trade_summary.Trade_duration.values)
            max_trade_duration = max(trade_summary.Trade_duration.values)
            
            backtest_results['all_trades']['avg_win']       = avg_win
            backtest_results['all_trades']['max_win']       = max_win
            backtest_results['all_trades']['avg_loss']      = avg_loss
            backtest_results['all_trades']['max_loss']      = max_loss
            backtest_results['all_trades']['win_rate']      = win_rate
            backtest_results['all_trades']['win_streak']    = longest_win_streak
            backtest_results['all_trades']['lose_streak']   = longest_lose_streak
            backtest_results['all_trades']['longest_trade'] = str(timedelta(seconds = int(max_trade_duration)))
            backtest_results['all_trades']['shortest_trade'] = str(timedelta(seconds = int(min_trade_duration)))
            backtest_results['all_trades']['avg_trade_duration'] = str(timedelta(seconds = int(avg_trade_duration)))
            backtest_results['all_trades']['net_pl']        = cpl.values[-1]
            
        # Cancelled and open orders
        cancelled_orders = self.broker.get_cancelled_orders(instrument)
        open_trades      = self.broker.get_open_positions(instrument)
        backtest_results['no_open'] = len(open_trades)
        backtest_results['no_cancelled'] = len(cancelled_orders)
        
        # Long trades
        long_trades     = trade_summary[trade_summary.Size > 0]
        no_long         = len(long_trades)
        backtest_results['long_trades'] = {}
        backtest_results['long_trades']['no_trades'] = no_long
        if no_long > 0:
            long_wins       = long_trades[long_trades.Profit > 0]
            avg_long_win    = np.mean(long_wins.Profit)
            max_long_win    = np.max(long_wins.Profit)
            long_loss       = long_trades[long_trades.Profit < 0]
            avg_long_loss   = abs(np.mean(long_loss.Profit))
            max_long_loss   = abs(np.min(long_loss.Profit))
            long_wr         = 100*len(long_trades[long_trades.Profit > 0])/no_long
            
            backtest_results['long_trades']['avg_long_win']     = avg_long_win
            backtest_results['long_trades']['max_long_win']     = max_long_win 
            backtest_results['long_trades']['avg_long_loss']    = avg_long_loss
            backtest_results['long_trades']['max_long_loss']    = max_long_loss
            backtest_results['long_trades']['long_wr']          = long_wr
            
          
        # Short trades
        short_trades    = trade_summary[trade_summary.Size < 0]
        no_short        = len(short_trades)
        backtest_results['short_trades'] = {}
        backtest_results['short_trades']['no_trades'] = no_short
        if no_short > 0:
            short_wins      = short_trades[short_trades.Profit > 0]
            avg_short_win   = np.mean(short_wins.Profit)
            max_short_win   = np.max(short_wins.Profit)
            short_loss      = short_trades[short_trades.Profit < 0]
            avg_short_loss  = abs(np.mean(short_loss.Profit))
            max_short_loss  = abs(np.min(short_loss.Profit))
            short_wr        = 100*len(short_trades[short_trades.Profit > 0])/no_short
            
            backtest_results['short_trades']['avg_short_win']   = avg_short_win
            backtest_results['short_trades']['max_short_win']   = max_short_win
            backtest_results['short_trades']['avg_short_loss']  = avg_short_loss
            backtest_results['short_trades']['max_short_loss']  = max_short_loss
            backtest_results['short_trades']['short_wr']        = short_wr
        
        return backtest_results
        
    
    def print_multibot_backtest_results(self, backtest_results=None):
        '''
        Prints to console the backtest results of a multi-bot backtest.
        
            Parameters:
                backtest_results (dict): dictionary containing backtest results.
        '''
        
        print("\n---------------------------------------------------")
        print("         MultiBot Backtest Results")
        print("---------------------------------------------------")
        print("Instruments traded: ", backtest_results.index.values)
        print("Final NAV:           ${}".format(round(self.broker.NAV, 2)))
        print("Total no trades:    ", backtest_results.no_trades.sum())
        print("Short trades:       ", backtest_results.no_short.sum(),
              "({}%)".format(round(100*backtest_results.no_short.sum()/backtest_results.no_trades.sum(),2)))
        print("Long trades:        ", backtest_results.no_long.sum(),
              "({}%)".format(round(100*backtest_results.no_long.sum()/backtest_results.no_trades.sum(),2)))
        print("\nInstrument win rates (%):")
        print(backtest_results[['win_rate', 'no_trades']])
        print("\nMaximum/Average Win/Loss breakdown ($):")
        print(backtest_results[["max_win", "max_loss", "avg_win", "avg_loss"]])
        print("\nAverage Reward to Risk Ratio:")
        print(round(backtest_results.avg_win / backtest_results.avg_loss,1))
        print("")
        
    
    def _assign_broker(self, broker_config):
        '''
        Configures and assigns appropriate broker for trading.
        '''
        
        if self.backtest_mode is True:
                utils_module    = importlib.import_module('autotrader.brokers.virtual.utils')
                
                utils           = utils_module.Utils()
                broker          = Broker(broker_config, utils)
                
                initial_deposit = self.backtest_initial_balance
                spread          = self.backtest_spread
                leverage        = self.backtest_leverage
                commission      = self.backtest_commission
                base_currency   = self.backtest_base_currency
                
                broker.make_deposit(initial_deposit)
                broker.fee      = spread
                broker.leverage = leverage
                broker.commission = commission
                broker.spread   = spread
                broker.base_currency = base_currency
                # self.get_data.base_currency = base_currency
                
                if int(self.verbosity) > 0:
                    banner = pyfiglet.figlet_format("AutoBacktest")
                    print(banner)
                
                
        else:
            
            if self.feed.lower() == 'yahoo':
                utils_module    = importlib.import_module('autotrader.brokers.virtual.utils')
                utils           = utils_module.Utils()
                broker          = Broker(broker_config, utils)
            else:
                utils_module    = importlib.import_module('autotrader.brokers.{}.utils'.format(self.feed.lower()))
                utils           = utils_module.Utils()
                broker          = Oanda.Oanda(broker_config, utils) # TODO - generalise away from Oanda
        
        self.broker = broker
        self.broker_utils = utils
    
    
    def _configure_emailing(self, global_config):
        '''
        Configure email settings.
        '''
        
        # TODO - allow setting email in this method
        
        if int(self.notify) > 0:
            host_email      = None
            mailing_list    = None
            
            # TODO - what if no email provided?
            
            if "EMAILING" in global_config:
                # Look for host email and mailing list in strategy config, if it
                # was not picked up in strategy config
                if "MAILING_LIST" in global_config["EMAILING"] and mailing_list is None:
                    mailing_list    = global_config["EMAILING"]["MAILING_LIST"]
                if "HOST_ACCOUNT" in global_config["EMAILING"] and host_email is None:
                    host_email      = global_config["EMAILING"]["HOST_ACCOUNT"]
            
            if host_email is None:
                print("Warning: email host account not provided.")
            if mailing_list is None:
                print("Warning: no mailing list provided.")
                
            email_params = {'mailing_list': mailing_list,
                            'host_email': host_email}
            self.email_params = email_params
            
            logfiles_path = os.path.join(self.home_dir, 'logfiles')
            order_summary_fp = os.path.join(logfiles_path, 'order_history.txt')
            
            if not os.path.isdir(logfiles_path):
                os.mkdir(logfiles_path)
            
            self.order_summary_fp = order_summary_fp


    def extract_backtest_results(self, trade_summary, broker, starting_balance, 
                                 utils):
        '''
        Analyses backtest summary to extract key statistics.
        '''
        backtest_results = {}
        
        # All trades
        no_trades   = len(trade_summary)
        backtest_results['no_trades'] = no_trades
        if no_trades > 0:
            backtest_results['all_trades'] = {}
            profit_abs  = broker.portfolio_balance - starting_balance
            profit_pc   = 100*profit_abs / starting_balance
            MDD         = round(broker.max_drawdown, 1)
            wins        = trade_summary[trade_summary.Profit > 0]
            avg_win     = np.mean(wins.Profit)
            max_win     = np.max(wins.Profit)
            loss        = trade_summary[trade_summary.Profit < 0]
            avg_loss    = abs(np.mean(loss.Profit))
            max_loss    = abs(np.min(loss.Profit))
            win_rate    = 100*broker.profitable_trades/no_trades
            longest_win_streak, longest_lose_streak  = utils.get_streaks(trade_summary)
            avg_trade_duration = np.mean(trade_summary.Trade_duration.values)
            min_trade_duration = min(trade_summary.Trade_duration.values)
            max_trade_duration = max(trade_summary.Trade_duration.values)
            
            backtest_results['all_trades']['profit_abs']    = profit_abs
            backtest_results['all_trades']['profit_pc']     = profit_pc
            backtest_results['all_trades']['MDD']           = MDD
            backtest_results['all_trades']['avg_win']       = avg_win
            backtest_results['all_trades']['max_win']       = max_win
            backtest_results['all_trades']['avg_loss']      = avg_loss
            backtest_results['all_trades']['max_loss']      = max_loss
            backtest_results['all_trades']['win_rate']      = win_rate
            backtest_results['all_trades']['win_streak']    = longest_win_streak
            backtest_results['all_trades']['lose_streak']   = longest_lose_streak
            backtest_results['all_trades']['longest_trade'] = str(timedelta(seconds = int(max_trade_duration)))
            backtest_results['all_trades']['shortest_trade'] = str(timedelta(seconds = int(min_trade_duration)))
            backtest_results['all_trades']['avg_trade_duration'] = str(timedelta(seconds = int(avg_trade_duration)))
        
        # Cancelled orders (insufficient margin)
        cancelled_orders    = broker.cancelled_orders
        backtest_results['no_open'] = len(broker.open_positions)
        backtest_results['no_cancelled'] = len(cancelled_orders)
        
        # Long trades
        long_trades     = trade_summary[trade_summary.Size > 0]
        no_long         = len(long_trades)
        backtest_results['long_trades'] = {}
        backtest_results['long_trades']['no_trades'] = no_long
        if no_long > 0:
            long_wins       = long_trades[long_trades.Profit > 0]
            avg_long_win    = np.mean(long_wins.Profit)
            max_long_win    = np.max(long_wins.Profit)
            long_loss       = long_trades[long_trades.Profit < 0]
            avg_long_loss   = abs(np.mean(long_loss.Profit))
            max_long_loss   = abs(np.min(long_loss.Profit))
            long_wr         = 100*len(long_trades[long_trades.Profit > 0])/no_long
            
            backtest_results['long_trades']['avg_long_win']     = avg_long_win
            backtest_results['long_trades']['max_long_win']     = max_long_win 
            backtest_results['long_trades']['avg_long_loss']    = avg_long_loss
            backtest_results['long_trades']['max_long_loss']    = max_long_loss
            backtest_results['long_trades']['long_wr']          = long_wr
            
          
        # Short trades
        short_trades    = trade_summary[trade_summary.Size < 0]
        no_short        = len(short_trades)
        backtest_results['short_trades'] = {}
        backtest_results['short_trades']['no_trades'] = no_short
        if no_short > 0:
            short_wins      = short_trades[short_trades.Profit > 0]
            avg_short_win   = np.mean(short_wins.Profit)
            max_short_win   = np.max(short_wins.Profit)
            short_loss      = short_trades[short_trades.Profit < 0]
            avg_short_loss  = abs(np.mean(short_loss.Profit))
            max_short_loss  = abs(np.min(short_loss.Profit))
            short_wr        = 100*len(short_trades[short_trades.Profit > 0])/no_short
            
            backtest_results['short_trades']['avg_short_win']   = avg_short_win
            backtest_results['short_trades']['max_short_win']   = max_short_win
            backtest_results['short_trades']['avg_short_loss']  = avg_short_loss
            backtest_results['short_trades']['max_short_loss']  = max_short_loss
            backtest_results['short_trades']['short_wr']        = short_wr
        
        # Save results
        self.backtest_results = backtest_results
        
        return backtest_results
    
    def print_backtest_results(self, backtest_results):
        # params      = self.strategy_params
        no_trades   = backtest_results['no_trades']
        if no_trades > 0:
            win_rate    = backtest_results['all_trades']['win_rate']
            profit_abs  = backtest_results['all_trades']['profit_abs']
            profit_pc   = backtest_results['all_trades']['profit_pc']
            MDD         = backtest_results['all_trades']['MDD']
            max_win     = backtest_results['all_trades']['max_win']
            avg_win     = backtest_results['all_trades']['avg_win']
            max_loss    = backtest_results['all_trades']['max_loss']
            avg_loss    = backtest_results['all_trades']['avg_loss']
            longest_win_streak = backtest_results['all_trades']['win_streak']
            longest_lose_streak = backtest_results['all_trades']['lose_streak']
        
        print("\n-------------------------------------------")
        print("            Backtest Results")
        print("-------------------------------------------")
        # TODO - the below are all strategy specific. Maybe if only one strategy
        # is used (ie len(self.strategies) = 1), that can be used. Otherwise,
        # not sure. However, the granularity has to be the same ... until 
        # time indexing becomes a thing
        # print("Strategy: {}".format(self.strategy.name))
        # print("Timeframe:               {}".format(params['granularity']))
        # if params is not None and 'RR' in params:
        #     print("Risk to reward ratio:    {}".format(params['RR']))
        #     print("Profitable win rate:     {}%".format(round(100/(1+params['RR']), 1)))
        if no_trades > 0:
            print("Backtest win rate:       {}%".format(round(win_rate, 1)))
            
            print("Total no. trades:        {}".format(self.broker.total_trades))
            print("Profit:                  ${} ({}%)".format(round(profit_abs, 3), 
                                              round(profit_pc, 1)))
            print("Maximum drawdown:        {}%".format(MDD))
            print("Max win:                 ${}".format(round(max_win, 2)))
            print("Average win:             ${}".format(round(avg_win, 2)))
            print("Max loss:                -${}".format(round(max_loss, 2)))
            print("Average loss:            -${}".format(round(avg_loss, 2)))
            print("Longest win streak:      {} trades".format(longest_win_streak))
            print("Longest losing streak:   {} trades".format(longest_lose_streak))
            print("Average trade duration   {}".format(backtest_results['all_trades']['avg_trade_duration']))
            
            
        else:
            print("No trades taken.")
        
        no_open = backtest_results['no_open']
        no_cancelled = backtest_results['no_cancelled']
        
        if no_open > 0:
            print("Orders still open:       {}".format(no_open))
        if no_cancelled > 0:
            print("Cancelled orders:        {}".format(no_cancelled))
        
        
        # Long trades
        no_long = backtest_results['long_trades']['no_trades']
        print("\n         Summary of long trades")
        print("-------------------------------------------")
        if no_long > 0:
            avg_long_win = backtest_results['long_trades']['avg_long_win']
            max_long_win = backtest_results['long_trades']['max_long_win']
            avg_long_loss = backtest_results['long_trades']['avg_long_loss']
            max_long_loss = backtest_results['long_trades']['max_long_loss']
            long_wr = backtest_results['long_trades']['long_wr']
            
            print("Number of long trades:   {}".format(no_long))
            print("Long win rate:           {}%".format(round(long_wr, 1)))
            print("Max win:                 ${}".format(round(max_long_win, 2)))
            print("Average win:             ${}".format(round(avg_long_win, 2)))
            print("Max loss:                -${}".format(round(max_long_loss, 2)))
            print("Average loss:            -${}".format(round(avg_long_loss, 2)))
        else:
            print("There were no long trades.")
          
        # Short trades
        no_short = backtest_results['short_trades']['no_trades']
        print("\n          Summary of short trades")
        print("-------------------------------------------")
        if no_short > 0:
            avg_short_win = backtest_results['short_trades']['avg_short_win']
            max_short_win = backtest_results['short_trades']['max_short_win']
            avg_short_loss = backtest_results['short_trades']['avg_short_loss']
            max_short_loss = backtest_results['short_trades']['max_short_loss']
            short_wr = backtest_results['short_trades']['short_wr']
            
            print("Number of short trades:  {}".format(no_short))
            print("short win rate:          {}%".format(round(short_wr, 1)))
            print("Max win:                 ${}".format(round(max_short_win, 2)))
            print("Average win:             ${}".format(round(avg_short_win, 2)))
            print("Max loss:                -${}".format(round(max_short_loss, 2)))
            print("Average loss:            -${}".format(round(avg_short_loss, 2)))
            
        else:
            print("There were no short trades.")


    def optimise(self, opt_params, bounds, Ns=4):
        '''
        Optimisation configuration.
        
            Parameters: 
                opt_params (list): the parameters to be optimised, as they 
                are named in the strategy configuration file.
                
                bounds (list of tuples): the bounds on each of the 
                parameters to be optimised, specified as a tuple of the form
                (lower, upper) for each parameter.
                
                Ns (int): the number of points along each dimension of the 
                optimisation grid.
                
        '''
        
        if type(bounds) == str:
            full_tuple = literal_eval(bounds)
            bounds = [(x[0], x[-1]) for x in full_tuple]

        if type(opt_params) == str:
            opt_params = opt_params.split(',')
        
        self.optimise_mode = True
        self.opt_params = opt_params
        self.bounds = bounds
        self.Ns = Ns
        
        
    def _run_optimise(self):
        '''
        Runs optimisation of strategy parameters.
        '''
        
        # Modify verbosity for optimisation
        verbosity = self.verbosity
        self.verbosity = 0
        self.show_plot = False
        
        self.objective      = 'profit + MDD'
        
        ''' --------------------------------------------------------------- '''
        '''                          Unpack user options                    '''
        ''' --------------------------------------------------------------- '''
        
        # Look in self.strategies for config
        if len(self.strategies) > 1:
            print("Error: please optimise one strategy at a time.")
            print("Exiting.")
            sys.exit(0)
        else:
            config_dict = self.strategies[list(self.strategies.keys())[0]]
                
        ''' --------------------------------------------------------------- '''
        '''                      Define optimisation inputs                 '''
        ''' --------------------------------------------------------------- '''
        my_args     = (config_dict, self.opt_params, self.verbosity)
        
        ''' --------------------------------------------------------------- '''
        '''                             Run Optimiser                       '''
        ''' --------------------------------------------------------------- '''
        start = timeit.default_timer()
        result = brute(func         = self._optimisation_helper_function, 
                       ranges       = self.bounds, 
                       args         = my_args, 
                       Ns           = self.Ns,
                       full_output  = True)
        stop = timeit.default_timer()
        
        ''' --------------------------------------------------------------- '''
        '''      Delete historical data file after running optimisation     '''
        ''' --------------------------------------------------------------- '''
        granularity             = config_dict["INTERVAL"]
        pair                    = config_dict["WATCHLIST"][0]
        historical_data_name    = 'hist_{0}{1}.csv'.format(granularity, pair)
        historical_quote_data_name = 'hist_{0}{1}_quote.csv'.format(granularity, pair)
        historical_data_file_path = os.path.join(self.home_dir, 
                                                 'price_data',
                                                 historical_data_name)
        historical_quote_data_file_path = os.path.join(self.home_dir, 
                                                       'price_data',
                                                       historical_quote_data_name)
        os.remove(historical_data_file_path)
        os.remove(historical_quote_data_file_path)
        
        opt_params = result[0]
        opt_value = result[1]
        
        # TODO - use the below for heatmap plotting
        # grid_points = result[2]
        # grid_values = result[3]
        
        ''' --------------------------------------------------------------- '''
        '''                           Print output                          '''
        ''' --------------------------------------------------------------- '''
        print("\nOptimisation complete.")
        print('Time to run: {}s'.format(round((stop - start), 3)))
        print("Optimal parameters:")
        print(opt_params)
        print("Objective:")
        print(opt_value)
        
        # Reset verbosity
        self.verbosity = verbosity
    
    
    def _optimisation_helper_function(self, params, config_dict, opt_params, verbosity):
        '''
        Helper function for optimising strategy parameters in AutoTrader.
        This function will parse the ordered params into the config dict.
        
        '''
        
        ''' ------------------------------------------------------------------ '''
        '''   Edit strategy parameters in config_dict using supplied params    '''
        ''' ------------------------------------------------------------------ '''
        for parameter in config_dict['PARAMETERS']:
            if parameter in opt_params:
                config_dict['PARAMETERS'][parameter] = params[opt_params.index(parameter)]
            else:
                continue
        
        ''' ------------------------------------------------------------------ '''
        '''           Run AutoTrader and evaluate objective function           '''
        ''' ------------------------------------------------------------------ '''
        self._clear_strategies()
        self._clear_bots()
        self.add_strategy(strategy_dict = config_dict)
        self._main()
        
        bot = self.bots_deployed[0]
        
            
        backtest_results    = self.analyse_backtest(bot.backtest_summary)
        
        try:
            objective           = -backtest_results['all_trades']['net_pl']
        except:
            objective           = 1000
                              
        print("Parameters/objective:", params, "/", objective)
        
        return objective
    
    

if __name__ == '__main__':
    autotrader = AutoTrader()
    autotrader.usage()



# -*- coding: utf-8 -*-
"""
Created on Sun May  2 17:18:38 2021

@author: kemac
"""

# Email
import smtplib, ssl
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from datetime import datetime
import pandas as pd

# Path management
import os



def send_order(order_details, mailing_list, host_email):
    '''
    Send order summary from oanda.
    '''
    
    instrument      = order_details['instrument']
    order_type      = order_details['order_type']
    size            = order_details['size']
    price           = order_details['order_price']
    stop_loss       = order_details['stop_loss']
    take_profit     = order_details['take_profit']
    strategy_name   = order_details['strategy']
    
    if instrument[-3:] == "JPY":
        val         = 2
    else:
        val         = 4
    
    if stop_loss is not None:
        stop_pips = round(abs(price - stop_loss)*10**val, 1)
        stop_loss = round(stop_loss, val+1)
    else:
        stop_pips = None
    
    if take_profit is not None:
        take_pips = round(abs(price - take_profit)*10**val, 1)
        take_profit = round(take_profit, val+1)
    else:
        take_pips = None
        
    # Email configuration settings
    sender_email    = host_email['email']
    password        = host_email['password']
    
    for person in mailing_list:
        
        last_name       = person.split('_')[1]
        title           = mailing_list[person]['title']
        receiver_email  = mailing_list[person]['email']
        
        # Constuct message details
        time        = datetime.now().strftime("%H:%M:%S")
        message     = MIMEMultipart("alternative")
        message["Subject"] = "{0} market order placed at {1}".format(instrument, 
                                                                     time)
        message["From"] = sender_email
        
        # Create the plain-text version of email
        plaintext       = ""
        
        # Load HTML version of email
        file_dir            = os.path.dirname(os.path.abspath(__file__))
        filename            = "{}_{}_{}.html".format(instrument, 
                                                    size,
                                                    datetime.now().timestamp())
        email_message_path  = os.path.join(file_dir, filename)
        
        # Write email in html
        with open(email_message_path, 'w+') as f:
            f.write('<p>Dear {} {},</p>\n'.format(title,
                                                  last_name))
            f.write('<p>A {} order has been placed for {} '.format(order_type, size))
            f.write('units of {} following an entry signal '.format(instrument))
            f.write('recieved by {}.</p>\n'.format(strategy_name))
            
            f.write('<p>A summary of the entry signal is provided below.</p>\n')
            
            f.write('<table border="1">\n')
            f.write('<tbody>\n')
            f.write('<tr>\n')
            f.write('<td>Pair</td>\n')
            f.write('<td>{}</td>\n'.format(instrument))
            f.write('</tr>\n')
            f.write('<tr>\n')
            f.write('<td>Size</td>\n')
            f.write('<td>{}</td>\n'.format(size))
            f.write('</tr>\n')
            f.write('<tr>\n')
            f.write('<td>Price</td>\n')
            f.write('<td>{}</td>\n'.format(price))
            f.write('</tr>\n')
            f.write('<tr>\n')
            f.write('<td>Time</td>\n')
            f.write('<td>{}</td>\n'.format(time))
            f.write('</tr>\n')
            f.write('<tr>\n')
            f.write('<td>Stop loss</td>\n')
            f.write('<td>{0} ({1})</td>\n'.format(stop_loss, stop_pips))
            f.write('</tr>\n')
            f.write('<tr>\n')
            f.write('<td>Take profit</td>\n')
            f.write('<td>{0} ({1})</td>\n'.format(take_profit, take_pips))
            f.write('</tr>\n')
            f.write('</tbody>\n')
            f.write('</table>\n')
            
            f.write('<hr />\n')
        

        email_body  = open(email_message_path, 'r').read()
        html        = email_body
        
        # Convert messages into plain/html MIMEText objects
        part1 = MIMEText(plaintext, "plain")
        part2 = MIMEText(html, "html")
        
        # Add HTML/plain-text parts to MIMEMultipart message
        message.attach(part1)
        message.attach(part2)
        
        # Create secure connection with server and send email
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL("smtp.gmail.com", 465, context=context) as server:
            server.login(sender_email, password)
            
            server.sendmail(sender_email, 
                            receiver_email, 
                            message.as_string()
                            )
            
        # Delete html file
        os.remove(email_message_path)


def send_order_summary(filepath, mailing_list, host_email):
    '''
    Send summary of orders placed with AutoTrader.
    '''
    
    file_dir        = os.path.dirname(os.path.abspath(__file__))
    order_history   = pd.read_csv(filepath, index_col=0, skipinitialspace=True)
    now             = datetime.now()
    date            = now.strftime("%B %d, %Y")
    
    # Email configuration settings
    sender_email    = host_email['email']
    password        = host_email['password']
    
    for person in mailing_list:
        last_name       = person.split('_')[1]
        title           = mailing_list[person]['title']
        receiver_email  = mailing_list[person]['email']
    
        # Constuct message details
        message         = MIMEMultipart("alternative")
        message["Subject"] = "AutoTrader Order Summary for {}".format(date)
        message["From"] = sender_email
        
        # Create the plain-text version of email
        plaintext       = ""
    
        # Create HTML email
        email_message_path  = os.path.join(file_dir, 'order_summary.html')
        
        # Write email in html
        with open(email_message_path, 'w+') as f:
            
            f.write('<p>Dear {} {},</p>\n'.format(title,
                                                  last_name))
            
            f.write('<p>Please find below a summary of orders placed on \n')
            f.write('your behalf by AutoTrader.</p>\n')
            
            f.write('<table border="1">\n')
            f.write('<tbody>\n')
            f.write('<tr>\n')
            f.write('<td>Order Time</td>\n')
            f.write('<td>Strategy</td>\n')
            # f.write('<td>Order Type</td>\n')  
            f.write('<td>Granularity</td>\n')  
            f.write('<td>Instrument</td>\n')
            f.write('<td>Signal Price</td>\n')
            f.write('<td>Size</td>\n')
            f.write('<td>Stop Loss</td>\n')
            f.write('<td>Take Profit</td>\n')
            f.write('</tr>\n')
            
            for index, row in order_history.iterrows():
                f.write('<tr>\n')
                f.write('<td>{}</td>\n'.format(index))
                f.write('<td>{}</td>\n'.format(row.strategy))
                # f.write('<td>{}</td>\n'.format(row.order_type))
                f.write('<td>{}</td>\n'.format(row.granularity))
                f.write('<td>{}/{}</td>\n'.format(row.instrument[:3], 
                                                  row.instrument[-3:]))
                f.write('<td>{}</td>\n'.format(row.trigger_price))
                f.write('<td>{}</td>\n'.format(row.order_size))
                f.write('<td>{}</td>\n'.format(round(row.stop_loss, 5)))
                f.write('<td>{}</td>\n'.format(round(row.take_profit, 5)))
                f.write('</tr>\n')
                
            f.write('</tbody>')
            f.write('</table>')
            
            # f.write('<p>&nbsp;</p>\n')
            f.write('<p>All the best in your trading endeavours,\n')
            f.write('<br />AutoTrader</p>\n')
            
        
        # Read file in
        email_body = open(email_message_path, 'r').read()
        html = email_body
            
        # Convert messages into plain/html MIMEText objects
        part1 = MIMEText(plaintext, "plain")
        part2 = MIMEText(html, "html")
        
        # Add HTML/plain-text parts to MIMEMultipart message
        message.attach(part1)
        message.attach(part2)
        
        # Create secure connection with server and send email
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL("smtp.gmail.com", 465, context=context) as server:
            server.login(sender_email, password)
    
            server.sendmail(sender_email, 
                            receiver_email, 
                            message.as_string()
                            )
        
        # Delete html file and order history file
        os.remove(email_message_path)
        os.remove(filepath)


def send_scan_results(scan_results, scan_details, mailing_list, host_email):
    '''
    Send results of market scan.
    
    '''
    time            = datetime.now().strftime("%H:%M:%S")
    index           = scan_details['index']
    strategy_name   = scan_details['strategy']
    
    # Email configuration settings
    sender_email    = host_email['email']
    password        = host_email['password']
    
    for person in mailing_list:
        # first_name      = person.split('_')[0]
        last_name       = person.split('_')[1]
        title           = mailing_list[person]['title']
        receiver_email  = mailing_list[person]['email']
        
        # Constuct message details
        message         = MIMEMultipart("alternative")
        message["Subject"] = "Scan Results for {} at {}".format(strategy_name, time)
        message["From"] = sender_email
        
        # Create the plain-text version of email
        plaintext       = ""
    
        # Load HTML version of email
        file_dir            = os.path.dirname(os.path.abspath(__file__))
        # file_dir = "/home/kieran/Documents/AT/development/AutoTrader/emailing/"
        email_message_path  = os.path.join(file_dir, 'scan_results.html')
        
        # Write email in html
        with open(email_message_path, 'w+') as f:
            if len(scan_results) > 0:
                f.write('<p>Dear {} {},</p>\n'.format(title,
                                                      last_name))
                f.write('<p>This is an automated message to notify you of \n')
                f.write('a recent match in a market scan you are running. The \n')
                f.write('details of the scan are as follows.\n')
                f.write('<br />Time of scan: {}.\n'.format(time))
                f.write('<br />Scan strategy: {}.\n'.format(strategy_name))
                f.write('<br />Scan index: {}.</p>\n'.format(index))
                
                f.write('<p>The results from the scan are shown in the table below.</p>\n')
                
                f.write('<table border="1">\n')
                f.write('<tbody>\n')
                f.write('<tr>\n')
                f.write('<td>Pair</td>\n')
                f.write('<td>Signal Price</td>\n')
                f.write('<td>Size</td>\n')
                f.write('<td>Stop Loss</td>\n')
                f.write('<td>Take Profit</td>\n')
                f.write('</tr>\n')
                
                
                for pair in scan_results:
                    size    = scan_results[pair]['size']
                    entry   = scan_results[pair]['entry'] 
                    stop    = scan_results[pair]['stop']
                    if stop is None:
                        stop = 'None'
                    else:
                        stop = round(stop, 5)

                    take    = scan_results[pair]['take']
                    if take is None:
                        take = 'None'
                    else:
                        take = round(take, 5)
                        
                    signal  = scan_results[pair]['signal']
                    
                    if size == 0:
                        size = 'Long' if signal == 1 else 'Short'
                    
                    f.write('<tr>\n')
                    f.write('<td>{}</td>\n'.format(pair))
                    f.write('<td>{}</td>\n'.format(round(entry, 5)))
                    f.write('<td>{}</td>\n'.format(size))
                    f.write('<td>{}</td>\n'.format(stop))
                    f.write('<td>{}</td>\n'.format(take))
                    f.write('</tr>\n')
                    
                f.write('</tbody>')
                f.write('</table>')
                
                f.write('<p>&nbsp;</p>\n')
                f.write('<p>All the best in your trading endeavours,</p>\n')
                f.write('<p>AutoTrader</p>\n')
            else:
                f.write('<p>Dear {} {},</p>\n'.format(title,
                                                      last_name))
                f.write('<p>This is an automated message to notify you \n')
                f.write('that the scan which you are running is still operational.\n')
                f.write('The details of the scan are as follows.</p>\n')
                f.write('<p>Time of scan: {}.</p>\n'.format(time))
                f.write('<p>Scan strategy: {}.</p>\n'.format(strategy_name))
                f.write('<p>Scan index: {}.</p>\n'.format(index))
                
                f.write('<p>All the best in your trading endeavours,\n')
                f.write('<br /><strong>AutoTrader</strong></p>\n')
            
        
        # Read file in
        email_body = open(email_message_path, 'r').read()
        html = email_body
            
        # Convert messages into plain/html MIMEText objects
        part1 = MIMEText(plaintext, "plain")
        part2 = MIMEText(html, "html")
        
        # Add HTML/plain-text parts to MIMEMultipart message
        message.attach(part1)
        message.attach(part2)
        
        # Create secure connection with server and send email
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL("smtp.gmail.com", 465, context=context) as server:
            server.login(sender_email, password)

            server.sendmail(
                            sender_email, 
                            receiver_email, 
                            message.as_string()
                            )
        
        # Delete html file
        os.remove(email_message_path)


def send_message(mailing_list, host_email, message):
    '''
    A method to email a generic message.
    
    Parameters:
        mailing_list (dict): a dictionary containing email contacts.
        
        host_email (dict): a dictionary containing the account details of the 
        host email account.
        
        message (str): the message to be sent.
    
    Refer to the AutoTrader documentation for more information:
    https://kieran-mackle.github.io/AutoTrader/docs/configuration-global#emailing
    '''

    # Email configuration settings
    sender_email = host_email['email']
    password = host_email['password']
    
    for person in mailing_list:
        receiver_email  = mailing_list[person]['email']

        # Create secure connection with server and send email
        context = ssl.create_default_context()
        with smtplib.SMTP_SSL("smtp.gmail.com", 465, context=context) as server:
            server.login(sender_email, password)
            server.sendmail(sender_email, 
                            receiver_email, 
                            message)
            


# -*- coding: utf-8 -*-
"""
Broker Utilities SuperClass

"""

import pandas as pd
import numpy as np
from datetime import datetime
import os


class BrokerUtils:
    
    def __init__(self):
        return
    
    def response_to_df(self, response):
        ''' Function to convert api response into a pandas dataframe. '''
        
        candles = response.body["candles"]
        times = []
        close_price, high_price, low_price, open_price = [], [], [], []
        
        for candle in candles:
            times.append(candle.time)
            close_price.append(float(candle.mid.c))
            high_price.append(float(candle.mid.h))
            low_price.append(float(candle.mid.l))
            open_price.append(float(candle.mid.o))
        
        dataframe = pd.DataFrame({"Open": open_price, "High": high_price, "Low": low_price, "Close": close_price})
        dataframe.index = pd.to_datetime(times)
        
        return dataframe
    
    
    def truncate(self, f, n):
        ''' Truncates a float f to n decimal places without rounding. '''
        s = '{}'.format(f)
        
        if 'e' in s or 'E' in s:
            return '{0:.{1}f}'.format(f, n)
        i, p, d = s.partition('.')
        
        return '.'.join([i, (d+'0'*n)[:n]])
    
    
    def get_pip_ratio(self, pair):
        ''' Function to return pip value ($/pip) of a given pair. '''
        if pair[-3:] == 'JPY':
            pip_value = 1e-2
        else:
            pip_value = 1e-4
        
        return pip_value
    
    
    def get_size(self, pair, amount_risked, price, stop_price, HCF, stop_distance = None):
        ''' Calculate position size based on account balance and risk profile. '''
        if stop_price is None and stop_distance is None:
            # No stop loss being used, instead risk portion of account
            units               = amount_risked/(HCF*price)
        else:
            pip_value           = self.get_pip_ratio(pair)
            
            if stop_price is None:
                pip_stop_distance = stop_distance
            else:
                pip_stop_distance = abs(price - stop_price) / pip_value
            
            
            if pip_stop_distance == 0:
                units           = 0
            else:
                quote_risk      = amount_risked / HCF
                price_per_pip   = quote_risk / pip_stop_distance
                units           = price_per_pip / pip_value
        
        return units
    
    
    def check_precision(self, pair, original_stop, original_take):
        ''' Modify stop/take based on pair for required ordering precision. ''' 
        if pair[-3:] == 'JPY':
            N = 3
        else:
            N = 5
        
        take_price      = float(self.truncate(original_take, N))
        stop_price      = float(self.truncate(original_stop, N))
        
        return stop_price, take_price
    
    
    def interval_to_seconds(self, interval):
        '''Converts the interval to time in seconds'''
        letter = interval[0]
        
        if len(interval) > 1:
            number = float(interval[1:])
        else:
            number = 1
        
        conversions = {'S': 1,
                       'M': 60,
                       'H': 60*60,
                       'D': 60*60*24
                       }
        
        my_int = conversions[letter] * number
        
        return my_int
    
    def write_to_order_summary(self, order_details, filepath):
        ''' Writes order details to summary file. '''
        
        # Check if file exists already, if not, create
        if not os.path.exists(filepath):
            f = open(filepath, "w")
            f.write("order time, strategy, granularity, order_type, instrument, order_size, ")
            f.write("trigger_price, stop_loss, take_profit\n")
            f.close()
        
        order_time          = order_details["order_time"]
        strategy            = order_details["strategy"]
        order_type          = order_details["order_type"]
        instrument          = order_details["instrument"]
        size                = order_details["size"]
        trigger_price       = order_details["order_price"]
        stop_loss           = order_details["stop_loss"]
        take_profit         = order_details["take_profit"]
        granularity         = order_details["granularity"]
        
        f                   = open(filepath, "a")
        f.write("{}, {}, {}, {}, {}, {}, {}, {}, {}\n".format(order_time, strategy, 
                                                              granularity, order_type, 
                                                              instrument, size, 
                                                              trigger_price, stop_loss, 
                                                              take_profit))
        f.close()
        
    
    def check_dataframes(self, df_1, df_2):
        '''Checks dataframe lengths and corrects if necessary'''
        
        if len(df_1) < len(df_2):
            new_df_1 = self.fix_dataframe(df_2, df_1)
            new_df_2 = df_2
        elif len(df_1) > len(df_2):
            new_df_2 = self.fix_dataframe(df_1, df_2)
            new_df_1 = df_1
            
        else:
            new_df_1 = df_1
            new_df_2 = df_2
        
        return new_df_1, new_df_2
    
    def fix_dataframe(self, df1, df2):
        ''' 
        Makes sure that the quote data and data dataframes are the same
        lenght.
        '''
        # Would be good to check which one is shorter, which is longer, then 
        # return both with corrections
        
        i1 = list(df1.index)
        i2 = list(df2.index)
        new_indices = list(set(i1) - set(i2))
        
        new_df = df2
        
        for index in new_indices:
            
            df_row = df1[df1.index == index]
            
            df_row.Open = None
            df_row.High = None
            df_row.Low = None
            df_row.Close = None
            
            new_df = new_df.append(df_row)
        
        new_df = new_df.sort_index()
        new_df = new_df.interpolate()
        
        return new_df
    
    def trade_summary(self, pair, closed_positions_dict):
        ''' Creates backtest trade summary dataframe. '''
        # Could also include a pairs list, and only input closed_pos dict
        order_ID    = []
        times_list  = []
        order_price = []
        entry_time  = []
        entry_price = []
        size        = []
        stop_price  = []
        take_price  = []
        profit      = []
        portfolio_balance = []
        exit_times  = []
        exit_prices = []
        trade_duration = []
        
        for order in closed_positions_dict:
            if closed_positions_dict[order]['instrument'] == pair:
                order_ID.append(closed_positions_dict[order]['order_ID'])
                entry_time.append(closed_positions_dict[order]['time_filled'])
                times_list.append(closed_positions_dict[order]['order_time'])
                order_price.append(closed_positions_dict[order]['order_price'])
                entry_price.append(closed_positions_dict[order]['entry_price'])
                size.append(closed_positions_dict[order]['size'])
                stop_price.append(closed_positions_dict[order]['stop_loss'])
                take_price.append(closed_positions_dict[order]['take_profit'])
                profit.append(closed_positions_dict[order]['profit'])
                portfolio_balance.append(closed_positions_dict[order]['balance'])
                exit_times.append(closed_positions_dict[order]['exit_time'])
                exit_prices.append(closed_positions_dict[order]['exit_price'])
                if type(closed_positions_dict[order]['exit_time']) == str:
                    exit_dt     = datetime.strptime(closed_positions_dict[order]['exit_time'],
                                                    "%Y-%m-%d %H:%M:%S%z")
                    entry_dt    = datetime.strptime(closed_positions_dict[order]['time_filled'],
                                                    "%Y-%m-%d %H:%M:%S%z")
                    trade_duration.append(exit_dt.timestamp() - entry_dt.timestamp())
                else:
                    trade_duration.append(closed_positions_dict[order]['exit_time'].timestamp() - 
                                          closed_positions_dict[order]['time_filled'].timestamp())
                
        dataframe = pd.DataFrame({"Order_ID": order_ID, 
                                  "Order_price": order_price,
                                  "Order_time": times_list,
                                  "Entry_time": entry_time,
                                  "Entry": entry_price, "Size": size,
                                  "Stop_loss": stop_price, "Take_profit": take_price,
                                  "Profit": profit, "Balance": portfolio_balance,
                                  "Exit_time": exit_times, "Exit_price": exit_prices,
                                  "Trade_duration": trade_duration})
        dataframe.index = pd.to_datetime(entry_time)
        dataframe = dataframe.sort_index()
        
        return dataframe
    
    def cancelled_order_summary(self, pair, positions_dict):
        ''' Creates cancelled order summary dataframe. '''
        order_ID    = []
        times_list  = []
        order_price = []
        size        = []
        stop_price  = []
        take_price  = []
        
        for order in positions_dict:
            if positions_dict[order]['instrument'] == pair:
                order_ID.append(positions_dict[order]['order_ID'])
                times_list.append(positions_dict[order]['order_time'])
                order_price.append(positions_dict[order]['order_price'])
                size.append(positions_dict[order]['size'])
                stop_price.append(positions_dict[order]['stop_loss'])
                take_price.append(positions_dict[order]['take_profit'])
                
        dataframe = pd.DataFrame({"Order_ID": order_ID, 
                                  "Order_price": order_price,
                                  "Size": size,
                                  "Stop_loss": stop_price, 
                                  "Take_profit": take_price})
        dataframe.index = pd.to_datetime(times_list)
        dataframe = dataframe.sort_index()
        
        return dataframe
        
    def open_order_summary(self, pair, positions_dict):
        ''' Creates open order summary dataframe. '''
        order_ID    = []
        times_list  = []
        order_price = []
        size        = []
        stop_price  = []
        take_price  = []
        entry_time  = []
        entry_price = []
        
        for order in positions_dict:
            if positions_dict[order]['instrument'] == pair:
                order_ID.append(positions_dict[order]['order_ID'])
                times_list.append(positions_dict[order]['order_time'])
                order_price.append(positions_dict[order]['order_price'])
                size.append(positions_dict[order]['size'])
                stop_price.append(positions_dict[order]['stop_loss'])
                take_price.append(positions_dict[order]['take_profit'])
                entry_time.append(positions_dict[order]['time_filled'])
                entry_price.append(positions_dict[order]['entry_price'])
                
                
        dataframe = pd.DataFrame({"Order_ID": order_ID, 
                                  "Order_price": order_price,
                                  "Order_time": times_list,
                                  "Size": size,
                                  "Stop_loss": stop_price, 
                                  "Take_profit": take_price,
                                  "Entry_time": entry_time,
                                  "Entry": entry_price})
        dataframe.index = pd.to_datetime(entry_time)
        dataframe = dataframe.sort_index()
        
        return dataframe
    
    def get_streaks(self, trade_summary):
        ''' Calculates longest winning and losing streaks from trade summary. '''
        profit_list             = trade_summary.Profit.values
        longest_winning_streak  = 1
        longest_losing_streak   = 1
        streak                  = 1
        
        for i in range(1, len(profit_list)):
            if np.sign(profit_list[i]) == np.sign(profit_list[i-1]):
                streak += 1
                
                if np.sign(profit_list[i]) > 0:
                    # update winning streak
                    longest_winning_streak  = max(longest_winning_streak, streak)
                else:
                    # Update losing 
                    longest_losing_streak   = max(longest_losing_streak, streak)
    
            else:
                streak = 1
        
        return longest_winning_streak, longest_losing_streak


    def reconstruct_portfolio(self, initial_balance, trade_summary, time_index):
        ''' REDUNDANT '''
        a = trade_summary.Exit_time.values
        
        profit_list = trade_summary.Profit.values.tolist()
        new_df = pd.DataFrame({"Profit": profit_list})
        new_df.index = pd.to_datetime(a, utc=True)
        
        balance   = initial_balance
        portfolio = []
        
        for timestamp in time_index:
            if timestamp in new_df.index:
                # Check if timestamp appears in multiple closed positions
                profit_vals = new_df.Profit[new_df.index == timestamp].values
                profit = 0
                for trade_profit in profit_vals:
                    profit += trade_profit
                
            else:
                profit = 0
            
            balance += profit
            portfolio.append(balance)
        
        dataframe = pd.DataFrame({"Balance": portfolio})
        dataframe.index = pd.to_datetime(time_index)
        dataframe = dataframe.sort_index()
        
        return dataframe


# -*- coding: utf-8 -*-
"""
Custom indicators library.
"""

from finta import TA
import numpy as np
import pandas as pd


def supertrend(data, period = 10, ATR_multiplier = 3.0, source=None):
    ''' Based on the SuperTrend indicator by KivancOzbilgic on TradingView '''
    
    if source is None:
        source = (data.High.values + data.Low.values) / 2
    
    # Calculate ATR
    atr             = TA.ATR(data, period)
    
    up              = source - (ATR_multiplier*atr)
    up_list         = [up[0]]
    up_times        = [data.index[0]]
    N_up            = 0
    
    dn              = source + (ATR_multiplier*atr)
    dn_list         = [dn[0]]
    dn_times        = [data.index[0]]
    N_dn            = 0
    
    trend           = 1
    trend_list      = [trend]
    
    for i in range(1, len(data)):
        
        if trend == 1:
            if data.Close.values[i] > max(up[N_up:i]):
                up_list.append(max(up[N_up:i]))
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
                
            else: 
                trend = -1
                N_dn = i
                dn_list.append(dn[i])
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
        else:
            if data.Close.values[i] < min(dn[N_dn:i]):
                dn_list.append(min(dn[N_dn:i]))
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
            else:
                trend = 1
                N_up = i
                up_list.append(up[i])
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
        
        trend_list.append(trend)
    
    # up_trend = pd.DataFrame({'uptrend': up_list}, index = up_times)
    # dn_trend = pd.DataFrame({'downtrend': dn_list}, index = dn_times)
    
    supertrend_df = pd.DataFrame({'uptrend': up_list,
                                  'downtrend': dn_list,
                                  'trend': trend_list}, 
                                 index = up_times)
    
    return supertrend_df


def stoch_rsi(data, K_period=3, D_period=3, RSI_length=14, Stochastic_length=14):

    rsi1 = TA.RSI(data, period=RSI_length)
    stoch = stochastic(rsi1, rsi1, rsi1, Stochastic_length)
    
    K = sma(stoch, K_period)
    D = sma(K, D_period)
    
    return K, D


def stochastic(high, low, close, period=14):
    
    K = np.zeros(len(high))
    
    for i in range(period, len(high)):
        low_val     = min(low[i-period+1:i+1])
        high_val    = max(high[i-period+1:i+1])
        
        K[i]        = 100 * (close[i] - low_val)/(high_val - low_val)
        
    return K
    

def sma(data, period=14):
    
    sma_list = []
    
    for i in range(len(data)):
        average = sum(data[i-period+1:i+1])/period
        sma_list.append(average)
    
    return sma_list


def ema(data, period=14, smoothing=2):
    
    ema = [sum(data[:period]) / period]
    
    for price in data[period:]:
        ema.append((price * (smoothing / (1 + period))) + ema[-1] * (1 - (smoothing / (1 + period))))
    
    for i in range(period-1):
        ema.insert(0, np.nan)
    
    return ema


def true_range(data, period=14):
    high_low = data['High'] - data['Low']
    high_close = np.abs(data['High'] - data['Close'].shift())
    low_close = np.abs(data['Low'] - data['Close'].shift())
    
    ranges = pd.concat([high_low, high_close, low_close], axis=1)
    true_range = np.max(ranges, axis=1)
    
    return true_range


def atr(data, period=14):
    
    tr = true_range(data, period)
    
    atr = tr.rolling(period).sum()/period
    
    return atr


def rsi():
    return

def crossover(list_1, list_2):
    ''' 
    Returns a list of length len(list_1) with 1 when list_1 crosses above
    list_2 and -1 when list_1 crosses below list_2.
    '''
    
    sign_list = []
    for i in range(len(list_1)):
        if np.isnan(list_1[i]):
            sign_list.append(np.nan)
        else:
            difference = list_1[i] - list_2[i]
            if difference < 0:
                sign_list.append(-1)
            else:
                sign_list.append(1)
    
    crossover_list = [0]
    
    for i in range(1, len(sign_list)):
        if sign_list[i] - sign_list[i-1] != 0:
            val = sign_list[i]
        else:
            val = 0
        
        crossover_list.append(val)

    return crossover_list


def cross_values(a, b, ab_crossover):
    cross_point_list = [0]
    last_cross_point = 0
    for i in range(1, len(ab_crossover)):
        if ab_crossover[i] != 0:
            i0 = 0
            m_a = a[i] - a[i-1]
            m_b = b[i] - b[i-1]
            ix = (b[i-1] - a[i-1])/(m_a-m_b) + i0
            
            cross_point = m_a*(ix - i0) + a[i-1]
            
            last_cross_point = cross_point
            
        else:
            cross_point = last_cross_point #0
        
        cross_point_list.append(cross_point)
    
    # Replace nans with 0
    cross_point_list = [0 if x!=x  else x for x in cross_point_list]
    
    return cross_point_list


def candles_between_crosses(cross_list):
    '''
    Returns candles since last cross
    
    
    Behaviour:
    in:  [0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1]
    out: [1, 2, 0, 1, 2, 3,  0, 1, 2, 3, 4, 5, 0]
    '''
    
    count = 0
    count_list = []
    
    for i in range(len(cross_list)):

        if cross_list[i] == 0:
            # Change in signal - reset count
            count += 1
        else:
            count = 0
        
        count_list.append(count)
    
    return count_list


def bullish_engulfing(data, detection = None):
    ''' Bullish engulfing pattern detection. '''
    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        
        down_trend  = np.where(data.Close.values < sma50, True, False)
        up_trend    = np.where(data.Close.values > sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    shadow_pc       = 100.0   # size of shadows 
    doji_pc         = 5.0
    shadow_factor   = 2.0 # number of times shadow dominates candle body
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ema(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    up_shadow       = data.High.values - body_high
    down_shadow     = body_low - data.Low.values
    
    has_up_shadow   = up_shadow > (shadow_pc / 100 * body)
    has_dn_shadow   = down_shadow > (shadow_pc / 100 * body)
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    candle_range    = data.High.values - data.Low.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    body_mid        = body/2 + body_low
    shadow_equals   = (up_shadow == down_shadow) | \
                          (
                              ((abs(up_shadow - down_shadow) / down_shadow * 100) < shadow_pc) & \
                              ((abs(down_shadow - up_shadow) / up_shadow * 100) < shadow_pc)
                          )
    doji_body       = (candle_range > 0) & (body <= candle_range * doji_pc / 100)
    doji            = doji_body & shadow_equals
    
    
    engulfing_bullish = [False]
    for i in range(1, len(data)):
        condition = down_trend[i] & \
                    white_body[i] & \
                    long_body[i] & \
                    black_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] >= data.Open.values[i-1]) & \
                    (data.Open.values[i] <= data.Close.values[i-1]) & \
                    ((data.Close.values[i] > data.Open.values[i-1]) | (data.Open.values[i] < data.Close.values[i-1]))
        
        engulfing_bullish.append(condition)
        
    return engulfing_bullish


def bearish_engulfing(data, detection = None):
    ''' Bearish engulfing pattern detection. '''
    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        
        down_trend  = np.where(data.Close.values < sma50, True, False)
        up_trend    = np.where(data.Close.values > sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    shadow_pc       = 100.0   # size of shadows 
    doji_pc         = 5.0
    shadow_factor   = 2.0 # number of times shadow dominates candle body
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ema(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    up_shadow       = data.High.values - body_high
    down_shadow     = body_low - data.Low.values
    
    has_up_shadow   = up_shadow > (shadow_pc / 100 * body)
    has_dn_shadow   = down_shadow > (shadow_pc / 100 * body)
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    candle_range    = data.High.values - data.Low.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    body_mid        = body/2 + body_low
    shadow_equals   = (up_shadow == down_shadow) | \
                          (
                              ((abs(up_shadow - down_shadow) / down_shadow * 100) < shadow_pc) & \
                              ((abs(down_shadow - up_shadow) / up_shadow * 100) < shadow_pc)
                          )
    doji_body       = (candle_range > 0) & (body <= candle_range * doji_pc / 100)
    doji            = doji_body & shadow_equals
    
    
    engulfing_bearish = [False]
    for i in range(1, len(data)):
        condition = up_trend[i] & \
                    black_body[i] & \
                    long_body[i] & \
                    white_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] <= data.Open.values[i-1]) & \
                    (data.Open.values[i] >= data.Close.values[i-1]) & \
                    ((data.Close.values[i] < data.Open.values[i-1]) | (data.Open.values[i] > data.Close.values[i-1]))
        
        engulfing_bearish.append(condition)
        
    return engulfing_bearish


def heikin_ashi(data):
    ''' 
        Calculates the Heikin-Ashi candlesticks from Japanese candlestick 
        data. 
    '''
    ha_close    = (data.Open + data.High + data.Low + data.Close)/4
    ha_open     = data.Open
    ha_low      = data.Low
    ha_high     = data.High
    
    for i in range(1, len(data)):
        ha_open[i]  = (data.Open[i-1] + data.Close[i-1])/2
        ha_low[i]   = min(data.Low[i], ha_open[i], ha_close[i])
        ha_high[i]  = max(data.High[i], ha_open[i], ha_close[i])
    
    ha_data = pd.concat([ha_open, ha_high, ha_low, ha_close], axis=1)
    ha_data.columns = ['Open', 'High', 'Low', 'Close']
    
    return ha_data
 

def find_swings(data, use_body = False):
    '''
        Locates the recent swings in price and returns a rolling list of 
        swing prices.
    '''
    
    # TODO - include resampling option to find swings on higher timeframe
    
    # use_body flag currently not implemented. Idea is to use body close/open
    # instead of high/low (wicks) of candle for swings.
    
    # oc2     = (data.Open.values + data.Close.values)/2
    hl2     = (data.High.values + data.Low.values)/2
    n       = 2
    EMA     = ema(hl2, n)
    
    grad    = np.gradient(EMA)
    
    swings   = np.zeros(len(grad))
    for i in range(1, len(grad)):
        if not np.isnan(grad[i-1]):
            if np.sign(grad[i]) != np.sign(grad[i-1]):
                swings[i] = -np.sign(grad[i])

    swing_df = pd.DataFrame(data=swings, index=data.index, columns=['swing'])
    
    low_list    = []
    high_list   = []
    for i in range(len(data)):
        if swing_df.swing[i] == -1:
            low_list.append(min(data.Low.values[i-1:i+2]))
            high_list.append(0)
        elif swing_df.swing[i] == 1:
            high_list.append(max(data.High.values[i-1:i+2]))
            low_list.append(0)
        else:
            low_list.append(0)
            high_list.append(0)
    
    swings_list     = merge_signals(low_list, high_list)
    last_swing      = rolling_signal_list(swings_list)
    last_swing[0:n] = list(data.High.values[0:n])
    
    # Need to return both a last swing low and last swing high list
    last_low        = rolling_signal_list(low_list)
    last_low[0:n]   = list(data.Low.values[0:n])
    last_high       = rolling_signal_list(high_list)
    last_high[0:n]  = list(data.High.values[0:n])
    
    swings          = pd.DataFrame(data={'Highs': last_high, 
                                         'Lows' : last_low,
                                         'Last' : last_swing},
                                   index = data.index)
    
    return swings
    

def rolling_signal_list(signals):
        ''' 
            Returns a list which maintains the previous signal, until a new 
            signal is given.
            
            [0,1,0,0,0,-1,0,0,1,0,0] ->  [0,1,1,1,1,-1,-1,-1,1,1,1]
            
        '''
        
        rolling_signals = [0]
        last_signal     = rolling_signals[0]
        
        for i in range(1, len(signals)):
            if signals[i] != 0:
                last_signal = signals[i]
            
            rolling_signals.append(last_signal)
        
        return rolling_signals

def unroll_signal_list(signals):
    ''' Unrolls a signal list. '''
    new_list = np.zeros(len(signals))
    
    for i in range(len(signals)):
        if signals[i] != signals[i-1]:
            new_list[i] = signals[i]
    
    return new_list


def merge_signals(signal_1, signal_2):
     ''' 
         Returns a single signal list which has merged two signal lists. 
     '''
     
     merged_signal_list = signal_1
     
     for i in range(len(signal_1)):
         if signal_2[i] != 0:
             merged_signal_list[i] = signal_2[i]
     
     return merged_signal_list


def ha_candle_run(ha_data):
    '''
        Returns a list for the number of consecutive green and red 
        Heikin-Ashi candles.
        
    '''
    green_candle    = np.where(ha_data.Close - ha_data.Open > 0, 1, 0)
    red_candle      = np.where(ha_data.Close - ha_data.Open < 0, 1, 0)
    
    green_run   = []
    red_run     = []
    
    green_sum   = 0
    red_sum     = 0
    
    for i in range(len(ha_data)):
        if green_candle[i] == 1:
            green_sum += 1
        else:
            green_sum = 0
        
        if red_candle[i] == 1:
            red_sum += 1
        else:
            red_sum = 0
        
        green_run.append(green_sum)
        red_run.append(red_sum)
        
    return green_run, red_run


def build_grid_price_levels(grid_origin, grid_space, grid_levels, 
                            grid_price_space=None, pip_value=0.0001):
    
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space*pip_value
    
    # Generate order_limit_price list 
    grid_price_levels = np.linspace(grid_origin - grid_levels*grid_price_space, 
                                     grid_origin + grid_levels*grid_price_space, 
                                     2*grid_levels + 1)
    
    return grid_price_levels

def build_grid(grid_origin, grid_space, grid_levels, order_direction, 
               order_type='stop-limit', grid_price_space=None, pip_value=0.0001, 
               take_distance=None, stop_distance=None, stop_type=None):
    '''
    grid_origin: origin of grid, specified as a price
    grid_space: spacing between grid levels, specified as pip distance
    grid_levels: number of grid levels either side of origin
    order_direction: the direction of each grid level order (1 for long, -1 for short)
    order_type: the order type of each grid level order
    '''
    # TODO - could add a limit price buffer, then use it to move the limit price
    # slightly away from the stop price
    
    # Check if stop_distance was provided without a stop_type
    if stop_distance is not None and stop_type is None:
        # set stop_type to 'limit' by default
        stop_type = 'limit'
    
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space*pip_value
    
    # Generate order_limit_price list 
    order_limit_prices = np.linspace(grid_origin - grid_levels*grid_price_space, 
                                     grid_origin + grid_levels*grid_price_space, 
                                     2*grid_levels + 1)
    
    # Construct nominal order
    nominal_order = {}
    nominal_order["order_type"]         = order_type
    nominal_order["direction"]          = order_direction
    nominal_order["stop_distance"]      = stop_distance
    nominal_order["stop_type"]          = stop_type
    nominal_order["take_distance"]      = take_distance
    
    # Build grid
    grid = {}

    for order, limit_price in enumerate(order_limit_prices):
        grid[order] = nominal_order.copy()
        grid[order]["order_stop_price"]  = order_limit_prices[order]
        grid[order]["order_limit_price"] = order_limit_prices[order]
        
    
    return grid

def merge_grid_orders(grid_1, grid_2):
    '''
    Merges grid dictionaries into one and re-labels order numbers so each
    order number is unique.
    '''
    # TODO - use **args/**kwargs to generalise how many grids are inputted
    order_offset = len(grid_1)
    grid = grid_1.copy()
    
    for order_no in grid_2:
        grid[order_no + order_offset] = grid_2[order_no]
    
    return grid
    
def last_level_crossed(data, base):
    ''' 
    Returns a list containing the last grid level touched.
    The grid levels are determined by the base input variable, 
    which corresponds to the pip_space x pip_value.
    '''
    # base = 20*0.0001
    
    last_level_crossed = np.nan
    levels_crossed = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]
        
        upper_prices = []
        lower_prices = []
        
        for price in [high, low]:    
            upper_prices.append(base*np.ceil(price/base))
            lower_prices.append(base*np.floor(price/base))
        
        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level
            last_level_crossed = lower_prices[0]
        
        levels_crossed.append(last_level_crossed)
    
    return levels_crossed


def build_multiplier_grid(origin, direction, multiplier, no_levels, precision, spacing):
    '''
    Constructs grid levels with a multiplying grid space.
    
        Parameters:
            origin (float): origin of grid as price amount.
            
            direction (int): direction of grid (1 for long, -1 for short).
            
            multiplier (float): grid space multiplier when price moves away 
            from the origin opposite to direction.
            
            no_levels (int): number of levels to calculate either side of the 
            origin.
            
            precision (int): instrument precision (eg. 4 for most currencies, 2 
            for JPY).
            
            spacing (float): spacing of grid in price units.
    '''
    
    levels = [i for i in range(1, no_levels + 1)]

    pos_levels = [round(origin + direction*spacing*i, precision) for i in levels]
    neg_spaces = [spacing*multiplier**(i) for i in levels]
    neg_levels = []
    prev_neg_level = origin
    for i in range(len(levels)):
        next_neg_level = prev_neg_level - direction*neg_spaces[i]
        prev_neg_level = next_neg_level
        neg_levels.append(round(next_neg_level, precision))
    
    grid = neg_levels + [origin] + pos_levels
    grid.sort()
    
    return grid


def last_level_touched(data, grid):
    '''
    Calculates the grid levels touched by price data.
    '''
    
    # initialise with nan
    last_level_crossed = np.nan 
    
    levels_touched = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]
        
        upper_prices = []
        lower_prices = []
        
        for price in [high, low]:    
            # Calculate level above
            upper_prices.append(grid[next(x[0] for x in enumerate(grid) if x[1] > price)])
            
            # calculate level below
            first_level_below_index = next(x[0] for x in enumerate(grid[::-1]) if x[1] < price)
            lower_prices.append(grid[-(first_level_below_index+1)])
        
        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level, since the level below the candle high
            # is different to the level below the candle low.
            # This essentially means the grid level is between candle low and high.
            last_level_crossed = lower_prices[0]
        
        levels_touched.append(last_level_crossed)
    
    return levels_touched


# def TDI(data):
#     rsiPeriod = input(11, minval = 1, title = "RSI Period")
#     bandLength = input(31, minval = 1, title = "Band Length")
#     lengthrsipl = input(1, minval = 0, title = "Fast MA on RSI")
#     lengthtradesl = input(9, minval = 1, title = "Slow MA on RSI")
    
#     src = close                                                             // Source of Calculations (Close of Bar)
#     r = rsi(src, rsiPeriod)                                                 // RSI of Close
#     ma = sma(r, bandLength)                                                 // Moving Average of RSI [current]
#     offs = (1.6185 * stdev(r, bandLength))                                  // Offset
#     up = ma + offs                                                          // Upper Bands
#     dn = ma - offs                                                          // Lower Bands
#     mid = (up + dn) / 2                                                     // Average of Upper and Lower Bands
#     fastMA = sma(r, lengthrsipl)                                            // Moving Average of RSI 2 bars back
#     slowMA = sma(r, lengthtradesl)                                          // Moving Average of RSI 7 bars back
    
#     hline(30)                                                               // Oversold
#     hline(50)                                                               // Midline
#     hline(70)                                                               // Overbought
    
#     upl = plot(up, "Upper Band", color = blue)                              // Upper Band
#     dnl = plot(dn, "Lower Band", color = blue)                              // Lower Band
#     midl = plot(mid, "Middle of Bands", color = orange, linewidth = 2)      // Middle of Bands
    
#     plot(slowMA, "Slow MA", color=green, linewidth=2)                       // Plot Slow MA
#     plot(fastMA, "Fast MA", color=red, linewidth=2)                         // Plot Fast MA
    
#     fill(upl, midl, red, transp=90)                                         // Fill Upper Half Red
#     fill(midl, dnl, green, transp=90)                                       // Fill Lower Half Green



